"""
THis module defines the DataSet class with is used to manipulate n
"""

from typing import Any, Callable, Optional

from deriva.core.ermrest_model import Model, Table

from deriva_ml.deriva_definitions import ML_SCHEMA


class Dataset:
    """
    Class to manipulate a dataset.

    Attributes:
        table: ERMrest table holding dataset information.
    """

    def __init__(self, model: Model):
        self._model = model
        self.ml_schema = ML_SCHEMA
        self._domain_schema = [
            s for s in model.schemas if s not in ["deriva-ml", "www", "public"]
        ].pop()
        self.table = self._model.schemas[self.ml_schema].tables["Dataset"]

    @staticmethod
    def export_dataset_element(
        spath: str, dpath: str, table: Table
    ) -> list[dict[str, Any]]:
        """Given a path in the data model, output an export specification for the path taken to get to the current table.

        Args:
          spath: Source path
          dpath: Destination path
          table: Table referenced to by the path

        Returns:
          The export specification that will retrieve that data from the catalog and place it into a BDBag.
        """
        # The table is the last element of the path.  Generate the ERMrest query by conversting the list of tables
        # into a path in the form of /S:T1/S:T2/S:Table
        # Generate the destination path in the file system using just the table names.

        exports = [
            {
                "source": {"api": "entity", "path": spath},
                "destination": {"name": dpath, "type": "csv"},
            }
        ]

        # If this table is an asset table, then we need to output the files associated with the asset.
        asset_columns = {"Filename", "URL", "Length", "MD5", "Description"}
        if asset_columns.issubset({c.name for c in table.columns}):
            exports.append(
                {
                    "source": {
                        "api": "attribute",
                        "path": f"{spath}/!(URL::null::)/url:=URL,length:=Length,filename:=Filename,md5:=MD5",
                    },
                    "destination": {"name": f"asset/{table.name}", "type": "fetch"},
                }
            )
        return exports

    def download_dataset_element(
        self, spath, dpath, table: Table
    ) -> list[dict[str, Any]]:
        """Return the download specification for the data object indicated by a path through the data model.

        Args:
          spath: Source path
          dpath: Destination path
          table: Table referenced to by the path

        Returns:
          The download specification that will retrieve that data from the catalog and place it into a BDBag.
        """
        exports = [
            {
                "processor": "csv",
                "processor_params": {
                    "query_path": f"/entity/{spath}?limit=none",
                    "output_path": dpath,
                },
            }
        ]

        # If this table is an asset table, then we need to output the files associated with the asset.
        asset_columns = {"Filename", "URL", "Length", "MD5", "Description"}
        if asset_columns.issubset({c.name for c in table.columns}):
            exports.append(
                {
                    "processor": "fetch",
                    "processor_params": {
                        "query_path": f"/attribute/{spath}/!(URL::null::)/url:=URL,length:=Length,filename:=Filename,md5:=MD5?limit=none",
                        "output_path": f"asset/{table.name}",
                    },
                }
            )
        return exports

    @staticmethod
    def _is_vocabulary(t: Table) -> bool:
        """

        Args:
          t: Table:

        Returns:
            True if the table has a vocabulary, False otherwise.
        """
        vocab_columns = {"Name", "URI", "Synonyms", "Description", "ID"}
        return vocab_columns.issubset({c.name for c in t.columns}) and t

    def _vocabulary_specification(
        self, writer: Callable[[str, str, Table], list[dict[str, Any]]]
    ) -> list[dict[str, Any]]:
        """

        Args:
          writer: Callable[[list[Table]]: list[dict[str: Any]]]:

        Returns:

        """
        vocabs = [
            table
            for s in self._model.schemas.values()
            for table in s.tables.values()
            if self._is_vocabulary(table)
        ]
        return [
            o
            for table in vocabs
            for o in writer(f"{table.schema.name}:{table.name}", table.name, table)
        ]

    def _domain_table_paths(
        self,
        graph: dict[Table, list[dict[Table, Any]]],
        spath: str = None,
        dpath: str = None,
        sprefix: str = "deriva-ml:Dataset/RID={Dataset_RID}",
        dprefix: str = "Dataset",
    ) -> list[tuple[str, str, Table]]:
        """Recursively walk over the domain schema graph and extend the current path.

        Args:
            graph: An undirected, acyclic graph of schema.  Represented as a dictionary whose name is the table name.
                and whose values are the child nodes of the table.
            spath: Source path so far
            dpath: Destination path so far
            sprefix: Initial path to be included.  Allows for nested datasets
            dprefix: Initial path to be included.  Allows for nested datasets

        Returns:
          A list of all the paths through the graph.  Each path is a list of tables.

        """
        source_path = spath or sprefix
        dest_path = dpath or dprefix
        paths = []
        for node, children in graph.items():
            if node.name == "Dataset":
                new_spath = sprefix
                new_dpath = dprefix
            else:
                new_spath = source_path + f"/{node.schema.name}:{node.name}"
                new_dpath = dest_path + f"/{node.name}"
            paths.append((new_spath, new_dpath, node))
            for child in children:
                paths.extend(self._domain_table_paths(child, new_spath, new_dpath))
        return paths

    def _table_paths(self, graph) -> list[tuple[str, str, Table]]:
        sprefix, dprefix = (
            "deriva-ml:Dataset/RID={Dataset_RID}",
            "Dataset",
        )
        table_paths = self._domain_table_paths(graph, sprefix=sprefix, dprefix=dprefix)
        dataset_dataset_table = self._model.schemas[self.ml_schema].tables['Dataset_Dataset']
        nested_sprefix = sprefix
        nested_dprefix = dprefix
        for i in range(1, 3):
        #    nested_sprefix += f'/DD{i}:=deriva-ml:Dataset_Dataset/D{i+1}:=(Nested_Dataset)=(deriva-ml:Dataset:RID)'
            nested_sprefix += f'/(RID)=(deriva-ml:Dataset_Dataset:Dataset)'
            nested_dprefix += f'/Dataset_Dataset'
            table_paths.append((nested_sprefix, nested_dprefix, dataset_dataset_table))
            nested_sprefix += f'/(Nested_Dataset)=(deriva-ml:Dataset:RID)'
            nested_dprefix += f'/Dataset'
            table_paths.append((nested_sprefix, nested_dprefix, self.table))
        # Get CSV for nested datasets.
            table_paths.extend(self._domain_table_paths(graph, sprefix=nested_sprefix, dprefix=nested_dprefix)[1:])
        return table_paths

    def _dataset_nesting_depth(self) -> int:
        ds_path = (
            self._model.catalog.getPathBuilder()
            .schemas[ML_SCHEMA]
            .tables["Dataset_Dataset"]
        )
        dsets = list(
            ds_path.attributes(ds_path.Dataset, ds_path.Nested_Dataset).fetch()
        )
        tree_depth = 3
        return 2

    def _schema_graph(
        self, node: Table, visited_nodes: Optional[set] = None
    ) -> dict[Table, list[dict[Table, list]]]:
        """Generate an undirected, acyclic graph of domain schema. We do this by traversing the schema foreign key
        relationships.  We stop when we hit the deriva-ml schema or when we reach a node that we have already seen.

        Nested datasets need to be unfolded

        Args:
          node: Current (starting) node in the graph.
          visited_nodes: param nested_dataset: Are we in a nested dataset, (i.e. have we seen the DataSet table)?

        Returns:
            Graph of the schema, starting from node.
        """

        visited_nodes = visited_nodes or set()
        graph = {node: []}

        def include_node(child: Table) -> bool:
            """Indicate if the table should be included in the graph.

            Include node in the graph if it's not a loopback from fk<-> referred_by, you have not already been to the
            node.
            """
            return (
                child != node
                and child not in visited_nodes
                and child.schema.name == self._domain_schema
            )

        # Get all the tables reachable from the end of the path avoiding loops from T1<->T2 via referenced_by
        nodes = {fk.pk_table for fk in node.foreign_keys if include_node(fk.pk_table)}
        nodes |= {fk.table for fk in node.referenced_by if include_node(fk.table)}

        for t in nodes:
            new_visited_nodes = visited_nodes.copy()
            new_visited_nodes.add(t)
            if self._is_vocabulary(t):
                # If the end of the path is a vocabulary table, we are at a terminal node in the ERD, so stop
                continue
            # Get all the paths that extend the current path
            graph[node].append(self._schema_graph(t, new_visited_nodes))
        return graph

    def _dataset_specification(
        self, writer: Callable[[str, str, Table], list[dict[str, Any]]]
    ) -> list[dict[str, Any]]:
        """Output a download/export specification for a dataset.  Each element of the dataset will be placed in its own dir
        The top level data directory of the resulting BDBag will have one subdirectory for element type. the subdirectory
        will contain the CSV indicating which elements of that type are present in the dataset, and then there will be a
        subdirectories for each object that is reachable from the dataset members.

        To simplify reconstructing the relationship between tables, the CVS for each
        The top level data directory will also contain a subdirectory for any controlled vocabularies used in the dataset.
        All assets will be placed into a directory named asset in a subdirectory with the asset table name.

        For example, consider a dataset that consists of two element types, T1 and T2. T1 has foreign key relationships to
        objects in tables T3 and T4.  There are also two controlled vocabularies, CV1 and CV2.  T2 is an asset table
        which has two asset in it. The layout of the resulting bdbag would be:
              data
                CV1/
                    cv1.csv
                CV2/
                    cv2.csv
                Dataset/
                    T1/
                        t1.csv
                        T3/
                            t3.csv
                        T4/
                            t4.csv
                    T2/
                        t2.csv
                asset/
                  T2
                    f1
                    f2

        Args:
          writer: Callable[[list[Table]]: list[dict[str:  Any]]]:

        Returns:
            A dataset specification.
        """
        element_spec = []
        for path in self._table_paths(self._schema_graph(self.table)):
            element_spec.extend(writer(*path))
        return self._vocabulary_specification(writer) + element_spec

    def export_outputs(self) -> list[dict[str, Any]]:
        """Return and output specification for the datasets in the provided model

        Returns:
          An export specification suitable for Chaise.
        """

        def writer(spath: str, dpath: str, table: Table) -> list[dict[str, Any]]:
            """

            Args:
              spath: list[Table]:
              dpath: list[Table]:
              table: Table

            Returns:
                An export specification suitable for Chaise.
            """
            return self.export_dataset_element(spath, dpath, table)

        # Export specification is a specification for the datasets, plus any controlled vocabulary
        return [
            {
                "source": {"api": False, "skip_root_path": True},
                "destination": {"type": "env", "params": {"query_keys": ["snaptime"]}},
            },
            {
                "source": {"api": "entity"},
                "destination": {
                    "type": "env",
                    "params": {"query_keys": ["RID", "Description"]},
                },
            },
            {
                "source": {"api": "schema", "skip_root_path": True},
                "destination": {"type": "json", "name": "schema"},
            },
        ] + self._dataset_specification(writer)

    def _processor_params(self) -> list[dict[str, Any]]:
        """
        Returns:
          a download specification for the datasets in the provided model.

        """

        def writer(spath: str, dpath: str, table: Table) -> list[dict[str, Any]]:
            """

            Args:
              spath:
              dpath:
              table: Table

            Returns:

            """
            return self.download_dataset_element(spath, dpath, table)

        # Download spec is the spec for any controlled vocabulary and for the dataset.
        return [
            {
                "processor": "json",
                "processor_params": {"query_path": f"/schema", "output_path": "schema"},
            }
        ] + self._dataset_specification(writer)

    def generate_dataset_download_spec(self) -> dict[str, Any]:
        """

        Returns:
        """
        return {
            "bag": {
                "bag_name": "Dataset_{Dataset_RID}",
                "bag_algorithms": ["md5"],
                "bag_archiver": "zip",
                "bag_metadata": {},
                "bag_idempotent": True,
            },
            "catalog": {
                "host": f"{self._model.catalog.deriva_server.scheme}://{self._model.catalog.deriva_server.server}",
                "catalog_id": f"{self._model.catalog.catalog_id}",
                "query_processors": [
                    {
                        "processor": "env",
                        "processor_params": {
                            "query_path": "/",
                            "output_path": "Dataset",
                            "query_keys": ["snaptime"],
                        },
                    }
                ]
                + self._processor_params(),
            },
        }



import sys
import argparse
from deriva.core import get_credential, ErmrestCatalog
from deriva_ml.schema_setup.annotations import generate_annotation


def alter_table_annotation(catalog: ErmrestCatalog, schema_name: str, table_name: str, annotation: dict):
    model_root = catalog.getCatalogModel()
    table = model_root.table(schema_name, table_name)
    table.alter(annotations=annotation)
    model_root.apply()


def alter_schema_annotation(catalog: ErmrestCatalog, schema_name: str, annotation: dict):
    model_root = catalog.getCatalogModel()
    schema = model_root.schemas[schema_name]
    schema.alter(annotations=annotation)
    model_root.apply()

def main():
    parser = argparse.ArgumentParser()
    parser.add_argument('--hostname', type=str, required=True)
    parser.add_argument('--schema_name', type=str, required=True)
    parser.add_argument('--catalog_id', type=str, required=True)
    args = parser.parse_args()
    credentials = get_credential(args.hostname)
    catalog = ErmrestCatalog('https', args.hostname, args.catalog_id, credentials)

    annotations = generate_annotation(args.schema_name, args.catalog_id)
    alter_table_annotation(catalog, args.schema_name, 'Workflow', annotations["workflow_annotation"])
    alter_table_annotation(catalog, args.schema_name, 'Execution', annotations["execution_annotation"])
    alter_table_annotation(catalog, args.schema_name, 'Execution_Metadata', annotations["execution_metadata_annotation"])
    alter_table_annotation(catalog, args.schema_name, 'Execution_Asset', annotations["execution_asset_annotation"])
    alter_schema_annotation(catalog, args.schema_name, annotations["schema_annotation"])

if __name__ == "__main__":
    sys.exit(main())



import argparse
import sys

from deriva.core.utils.core_utils import tag as deriva_tags
from ..upload import bulk_upload_configuration

def generate_annotation(catalog_id: str, schema: str) -> dict:
    workflow_annotation = {
        deriva_tags.visible_columns: {
            "*": [
                "RID",
                "Name",
                "Description",
                {"display": {"markdown_pattern": "[{{{URL}}}]({{{URL}}})"}, "markdown_name": "URL"},
                "Checksum",
                "Version",
                {"source": [{"outbound": [schema, "Workflow_Workflow_Type_fkey"]}, "RID"]}
            ]
        }
    }

    execution_annotation = {
        deriva_tags.visible_columns: {
            "*": [
                "RID",
                [schema, "Execution_RCB_fkey"],
                "RCT",
                "Description",
                {"source": [{"outbound": [schema, "Execution_Workflow_fkey"]}, "RID"]},
                "Duration",
                "Status",
                "Status_Detail"
            ]
        },
        "tag:isrd.isi.edu,2016:visible-foreign-keys": {
            "detailed": [
                {
                    "source": [{"inbound": [schema, "Dataset_Execution_Execution_fkey"]},
                               {"outbound": [schema, "Dataset_Execution_Dataset_fkey"]}, "RID"],
                    "markdown_name": "Dataset"
                },
                {
                    "source": [
                        {"inbound": [schema, "Execution_Asset_Execution_Execution_fkey"]},
                        {"outbound": [schema, "Execution_Asset_Execution_Execution_Asset_fkey"]}, "RID"],
                    "markdown_name": "Execution Asset"
                },
                {
                    "source": [{"inbound": [schema, "Execution_Metadata_Execution_fkey"]}, "RID"],
                    "markdown_name": "Execution Metadata"
                }
            ]
        }
    }

    execution_asset_annotation = {
        deriva_tags.table_display: {
            "row_name": {
                "row_markdown_pattern": "{{{Filename}}}"
            }
        },
        deriva_tags.visible_columns: {
            "compact": [
                "RID",
                "URL",
                "Description",
                "Length", [schema, "Execution_Asset_Execution_Asset_Type_fkey"],
                # {
                #     "display": {
                #         "template_engine": "handlebars",
                #         "markdown_pattern": "{{#if (eq  _Execution_Asset_Type \"2-5QME\")}}\n ::: iframe []("
                #                             "https://dev.eye-ai.org/~vivi/deriva-webapps/plot/?config=test-line"
                #                             "-plot&Execution_Asset_RID={{{RID}}}){class=chaise-autofill "
                #                             "style=\"min-width: 500px; min-height: 300px;\"} \\n:::\n {{/if}}"
                #     },
                #     "markdown_name": "ROC Plot"
                # }
            ],
            "detailed": [
                "RID",
                "RCT",
                "RMT",
                "RCB",
                "RMB",
                # {
                #     "display": {
                #         "template_engine": "handlebars",
                #         "markdown_pattern": "{{#if (eq _Execution_Asset_Type \"2-5QME\")}} ::: iframe []("
                #                             "https://dev.eye-ai.org/~vivi/deriva-webapps/plot/?config=test-line"
                #                             "-plot&Execution_Asset_RID={{{RID}}}){style=\"min-width:1000px; "
                #                             "min-height:700px; height:70vh;\" class=\"chaise-autofill\"} \\n::: {"
                #                             "{/if}}"
                #     },
                #     "markdown_name": "ROC Plot"
                # },
                "URL",
                "Filename",
                "Description",
                "Length",
                "MD5",
                [schema, "Execution_Asset_Execution_Asset_Type_fkey"]
            ]
        }
    }

    execution_metadata_annotation = {
        deriva_tags.table_display: {
            "row_name": {
                "row_markdown_pattern": "{{{Filename}}}"
            }
        }
    }

    dataset_annotation = {
        # Setup Facet on types
        # Make types in visible columns
        # Have all connected values be visabile FK.
    }

    schema_annotation = {
        'name_style': {'underline_space': True},
    }

    catalog_annotation = {
        deriva_tags.display: {"name_style": { "underline_space": True}},
        deriva_tags.chaise_config: {
            "headTitle": "Catalog ML",
            "navbarBrandText": "ML Data Browser",
            "systemColumnsDisplayEntry": ["RID"],
            "systemColumnsDisplayCompact": ["RID"],
            "navbarMenu": {
                "newTab": False,
                "children": [
                    {
                        "name": "User Info",
                        "children": [
                            {
                                "url": f"/chaise/recordset/#{catalog_id}/public:ERMrest_Client",
                                "name": "Users"
                            },
                            {
                                "url": f"/chaise/recordset/#{catalog_id}/public:ERMrest_Group",
                                "name": "Groups"
                            },
                            {
                                "url": f"/chaise/recordset/#{catalog_id}/public:ERMrest_RID_Lease",
                                "name": "ERMrest RID Lease"
                            }
                        ]
                    },
                    {
                        "name": "Deriva-ML",
                        "children": [
                            {
                                "url": f"/chaise/recordset/#{catalog_id}/{schema}:Workflow",
                                "name": "Workflow"
                            },
                            {
                                "url": f"/chaise/recordset/#{catalog_id}/{schema}:Workflow_Type",
                                "name": "Workflow Type"
                            },
                            {
                                "url": f"/chaise/recordset/#{catalog_id}/{schema}:Execution",
                                "name": "Execution"
                            },
                            {
                                "url": f"/chaise/recordset/#{catalog_id}/{schema}:Execution_Metadata",
                                "name": "Execution Metadata"
                            },
                            {
                                "url": f"/chaise/recordset/#{catalog_id}/{schema}:Execution_Metadata_Type",
                                "name": "Execution Metadata Type"
                            },
                            {
                                "url": f"/chaise/recordset/#{catalog_id}/{schema}:Execution_Asset",
                                "name": "Execution Asset"
                            },
                            {
                                "url": f"/chaise/recordset/#{catalog_id}/{schema}:Execution_Asset_Type",
                                "name": "Execution Asset Type"
                            },
                            {
                                "url": f"/chaise/recordset/#{catalog_id}/{schema}:Dataset",
                                "name": "Dataset"
                            }
                        ]
                    }
                ]
            },
            "defaultTable": {"table": "Dataset", "schema": "deriva-ml"},
            "deleteRecord": True,
            "showFaceting": True,
            "shareCiteAcls": True,
            "exportConfigsSubmenu": {"acls": {"show": ["*"], "enable": ["*"]}},
            "resolverImplicitCatalog": catalog_id,
        },
        deriva_tags.bulk_upload: bulk_upload_configuration,
    }

    return {"workflow_annotation": workflow_annotation,
            "dataset_annotation": dataset_annotation,
            "execution_annotation": execution_annotation,
            "execution_asset_annotation": execution_asset_annotation,
            "execution_metadata_annotation": execution_metadata_annotation,
            "schema_annotation": schema_annotation,
            "catalog_annotation": catalog_annotation,
            }


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument('--catalog_id', type=str, required=True)
    parser.add_argument('--schema_name', type=str, required=True)
    args = parser.parse_args()
    generate_annotation(args.catalog_id, args.schema_name)


if __name__ == "__main__":
    sys.exit(main())



"""
`deriva_ml_base.py` is the core module for the Deriva ML project.  This module implements the DeriaML class, which is
the primary interface to the Deriva based catalogs.  The module also implements the Feature and Vocabulary functions
in the DerivaML.

DerivaML and its associated classes all depend on a catalog that implements a `deriva-ml` schema with tables and
relationships that follow a specific data model.

"""

import getpass
import json
import logging
import os
import re
from collections import defaultdict
from datetime import datetime
from enum import Enum
from itertools import chain
from pathlib import Path
from tempfile import mkdtemp,  NamedTemporaryFile, TemporaryDirectory
from types import UnionType
from typing import Optional, Any, Iterable, Type, ClassVar, TYPE_CHECKING

import requests
from bdbag import bdbag_api as bdb
from bdbag.fetch.fetcher import fetch_single_file
from deriva.core import (
    ErmrestCatalog,
    get_credential,
    format_exception,
    urlquote,
    DEFAULT_SESSION_CONFIG,
)
from deriva.core.datapath import DataPathException, _ResultSet
from deriva.core.datapath import _CatalogWrapper
from deriva.core.ermrest_catalog import ResolveRidResult
from deriva.core.ermrest_model import FindAssociationResult, Column, Key, Table
from deriva.core.hatrac_store import HatracStore
from deriva.core.utils import hash_utils, mime_utils
from deriva.core.utils.hash_utils import compute_file_hashes
from deriva.transfer.download.deriva_download import GenericDownloader
from deriva.transfer.upload.deriva_upload import GenericUploader
from pydantic import BaseModel, Field, create_model, validate_call, ConfigDict

from .execution_configuration import ExecutionConfiguration
from .schema_setup.dataset_annotations import generate_dataset_annotations
from .upload import asset_dir
from .upload import table_path, bulk_upload_configuration
from .dataset import Dataset
from .deriva_definitions import ColumnDefinition
from .deriva_definitions import MLVocab, ExecMetadataVocab
from .deriva_definitions import (
    RID,
    UploadState,
    Status,
    FileUploadState,
    DerivaMLException,
    ML_SCHEMA,
)

if TYPE_CHECKING:
    from .execution import Execution


class VocabularyTerm(BaseModel):
    """An entry in a vocabulary table.

    Attributes:
       name: Name of vocabulary term
       synonyms: List of alternative names for the term
       id: CURI identifier for the term
       uri: Unique URI for the term.
       description: A description of the meaning of the term
       rid: Resource identifier assigned to the term

    Args:

    Returns:

    """

    name: str = Field(alias="Name")
    synonyms: Optional[list[str]] = Field(alias="Synonyms")
    id: str = Field(alias="ID")
    uri: str = Field(alias="URI")
    description: str = Field(alias="Description")
    rid: str = Field(alias="RID")

    class Config:
        """ """

        extra = "ignore"


class SemanticVersion(Enum):
    """Simple enumeration for semantic versioning."""

    major = "major"
    minor = "minor"
    patch = "patch"


class FeatureRecord(BaseModel):
    """Base class for feature records.  Feature records are pydantic models which are dynamically generated and
    describe all of the columns of a feature.

    Args:

    Returns:

    """

    # model_dump of this feature should be compatible with feature table columns.
    Execution: str
    Feature_Name: str
    feature: ClassVar[Optional["Feature"]] = None

    class Config:
        """ """

        arbitrary_types_allowed = True

    @classmethod
    def feature_columns(cls) -> set[Column]:
        """

        Args:

        Returns:
          :return: set of feature column names.

        """
        return cls.feature.feature_columns

    @classmethod
    def asset_columns(cls) -> set[Column]:
        """

        Args:

        Returns:
          A set of asset column names.

        """
        return cls.feature.asset_columns

    @classmethod
    def term_columns(cls) -> set[Column]:
        """

        Args:

        Returns:
          :return: set of term column names.

        """
        return cls.feature.term_columns

    @classmethod
    def value_columns(cls) -> set[Column]:
        """

        Args:

        Returns:
          A set of value column names.

        """
        return cls.feature.value_columns


class Feature:
    """Wrapper for results of Table.find_associations()"""

    def __init__(self, atable: FindAssociationResult):
        self.feature_table = atable.table
        self.target_table = atable.self_fkey.pk_table
        self.feature_name = atable.table.columns["Feature_Name"].default

        def is_asset(table):
            """

            Args:
              table:

            Returns:

            """
            asset_columns = {"Filename", "URL", "Length", "MD5", "Description"}
            return asset_columns.issubset({c.name for c in table.columns})

        def is_vocabulary(table):
            """

            Args:
              table:

            Returns:

            """
            vocab_columns = {"NAME", "URI", "SYNONYMS", "DESCRIPTION", "ID"}
            return vocab_columns.issubset({c.name.upper() for c in table.columns})

        skip_columns = {
            "RID",
            "RMB",
            "RCB",
            "RCT",
            "RMT",
            "Feature_Name",
            self.target_table.name,
            "Execution",
        }
        self.feature_columns = {
            c for c in self.feature_table.columns if c.name not in skip_columns
        }

        assoc_fkeys = {atable.self_fkey} | atable.other_fkeys

        # Determine the role of each column in the feature outside the FK columns.
        self.asset_columns = {
            fk.foreign_key_columns[0]
            for fk in self.feature_table.foreign_keys
            if fk not in assoc_fkeys and is_asset(fk.pk_table)
        }

        self.term_columns = {
            fk.foreign_key_columns[0]
            for fk in self.feature_table.foreign_keys
            if fk not in assoc_fkeys and is_vocabulary(fk.pk_table)
        }

        self.value_columns = self.feature_columns - (
                self.asset_columns | self.term_columns
        )

    def feature_record_class(self) -> type[FeatureRecord]:
        """ "Create a pydantic model for entries into the specified feature table

        Returns:
            A Feature class that can be used to create instances of the feature.
        """

        def map_type(c: Column) -> UnionType | Type[str] | Type[int] | Type[float]:
            """Map a deriva type into a pydantic model type.

            Args:
                c: column to be mapped
                c: Column:

            Returns:
                A pydantic model type
            """
            if c.name in {c.name for c in self.asset_columns}:
                return str | Path

            match c.type.typename:
                case "text":
                    return str
                case "int2" | "int4" | "int8":
                    return int
                case "float4" | "float8":
                    return float
                case _:
                    return str

        featureclass_name = f"{self.target_table.name}Feature{self.feature_name}"

        # Create feature class. To do this, we must determine the python type for each column and also if the
        # column is optional or not based on its nulliblity.
        feature_columns = {
                              c.name: (
                                  Optional[map_type(c)] if c.nullok else map_type(c),
                                  c.default or None,
                              )
                              for c in self.feature_columns
                          } | {
                              "Feature_Name": (
                                  str,
                                  self.feature_name,
                              ),  # Set default value for Feature_Name
                              self.target_table.name: (str, ...),
                          }
        docstring = f"Class to capture fields in a feature {self.feature_name} on table {self.target_table}. Feature columns include:\n"
        docstring += "\n".join([f"    {c.name}" for c in self.feature_columns])

        model = create_model(
            featureclass_name,
            __base__=FeatureRecord,
            __doc__=docstring,
            **feature_columns,
        )
        model.feature = (
            self  # Set value of class variable within the feature class definition.
        )

        return model

    def __repr__(self) -> str:
        return (
            f"Feature(target_table={self.target_table.name}, feature_name={self.feature_name}, "
            f"feature_table={self.feature_table.name})"
        )


class DerivaML:
    """Base class for ML operations on a Deriva catalog.

    This class is intended to be used as a base class on which more domain specific interfaces are built.

    Attributes:
        host_name: Hostname of the Deriva server.
        catalog_id: Catalog ID. Either and identifier, or a catalog name.
        domain_schema: Schema name for domain specific tables and relationships.
        model: ERMRest model for the catalog
    """

    def __init__(
            self,
            hostname: str,
            catalog_id: str | int,
            domain_schema: str = None,
            project_name: str = None,
            cache_dir: Optional[str] = None,
            working_dir: Optional[str] = None,
            model_version: str = "1",
            ml_schema: str = ML_SCHEMA,
            logging_level=logging.WARNING,
    ):
        """Create and initialize a DerivaML instance.

        This method will connect to a catalog, and initialize local configuration for the ML execution.
        This class is intended to be used as a base class on which domain-specific interfaces are built.

        Args:
            hostname: Hostname of the Deriva server.
            catalog_id: Catalog ID. Either and identifier, or a catalog name.
            domain_schema: Schema name for domain specific tables and relationships.
            cache_dir: Directory path for caching data downloaded from the Deriva server as bdbag.
            working_dir: Directory path for storing data used by or generated by any computations.
            model_version: A string that indicates the version model.  Typically passed in via
        """
        self.host_name = hostname
        self.catalog_id = catalog_id
        self.ml_schema = ml_schema
        self.version = model_version

        self.credential = get_credential(hostname)
        self.catalog = ErmrestCatalog(
            "https",
            hostname,
            catalog_id,
            self.credential,
            session_config=self._get_session_config(),
        )
        self.model = self.catalog.getCatalogModel()
        self.configuration = None
        self.dataset_table = self.model.schemas[self.ml_schema].tables["Dataset"]

        builtin_schemas = ["public", self.ml_schema, "www"]
        self.domain_schema = (
                domain_schema
                or [s for s in self.model.schemas.keys() if s not in builtin_schemas].pop()
        )
        self.project_name = project_name or self.domain_schema

        self.start_time = datetime.now()
        self.status = Status.pending.value
        tdir = None
        if cache_dir:
            self.cache_dir = Path(cache_dir)
            self.cache_dir.mkdir(parents=True, exist_ok=True)
        else:
            tdir = mkdtemp()
            self.cache_dir = Path(tdir)
        default_workdir = self.__class__.__name__ + "_working"
        if working_dir:
            self.working_dir = Path(working_dir).joinpath(
                getpass.getuser(), default_workdir
            )
        else:
            tdir = tdir or mkdtemp()
            self.working_dir = Path(tdir) / default_workdir
        self.working_dir.mkdir(parents=True, exist_ok=True)
        logging.basicConfig(
            level=logging_level, format="%(asctime)s - %(levelname)s - %(message)s"
        )
        if "dirty" in self.version:
            logging.info(
                f"Loading dirty model.  Consider commiting and tagging: {self.version}"
            )

    @staticmethod
    def _get_session_config():
        """ """
        session_config = DEFAULT_SESSION_CONFIG.copy()
        session_config.update(
            {
                # our PUT/POST to ermrest is idempotent
                "allow_retry_on_all_methods": True,
                # do more retries before aborting
                "retry_read": 8,
                "retry_connect": 5,
                # increase delay factor * 2**(n-1) for Nth retry
                "retry_backoff_factor": 5,
            }
        )
        return session_config

    @property
    def pathBuilder(self) -> _CatalogWrapper:
        """Get a new instance of a pathBuilder object."""
        return self.catalog.getPathBuilder()

    def dataset_version(self, dataset_rid: RID) -> tuple[int, ...]:
        """Retrieve the version of the specified dataset.

        Args:
            dataset_rid: return: A tuple with the semantic version of the dataset.
            dataset_rid: RID:

        Returns:
            A tuple with the semantic version of the dataset.
        """
        rid_record = self.resolve_rid(dataset_rid)
        if rid_record.table.name != self.dataset_table.name:
            raise DerivaMLException(
                f"RID: {dataset_rid} does not belong to dataset {self.dataset_table.name}"
            )
        return tuple(map(int, self.retrieve_rid(dataset_rid)["Version"].split(".")))

    def increment_dataset_version(
            self, dataset_rid: RID, component: SemanticVersion
    ) -> tuple[int, ...]:
        """Increment the version of the specified dataset.

        Args:
          dataset_rid: RID to a dataset
          component: Which version of the dataset to increment.
          dataset_rid: RID:
          component: SemanticVersion:

        Returns:
          new vsemantic ersion of the dataset as a 3-tuple

        Raises:
          DerivaMLException: if provided RID is not to a dataset.
        """
        major, minor, patch = self.dataset_version(dataset_rid)
        match component:
            case SemanticVersion.major:
                major += 1
            case SemanticVersion.minor:
                minor += 1
            case SemanticVersion.patch:
                patch += 1
        dataset_path = self.pathBuilder.schemas[self.ml_schema].tables[
            self.dataset_table.name
        ]
        dataset_path.update(
            [{"RID": dataset_rid, "Version": f"{major}.{minor}.{patch}"}]
        )
        return major, minor, patch

    @property
    def domain_path(self):
        """Get a new instance of a pathBuilder object to the domain schema"""

        return self.pathBuilder.schemas[self.domain_schema]

    def _get_table(self, table: str | Table) -> Table:
        """Return the table object corresponding to the given table name.

        If the table name appears in more than one schema, return the first one you find.

        Args:
          table: A ERMRest table object or a string that is the name of the table.
          table: str | Table:

        Returns:
          Table object.
        """
        if isinstance(table, Table):
            return table
        for s in self.model.schemas.values():
            if table in s.tables.keys():
                return s.tables[table]
        raise DerivaMLException(f"The table {table} doesn't exist.")

    def table_path(self, table: str | Table) -> Path:
        """Return a local file path in which to place a CSV to add values to a table on upload.

        Args:
          table: return:
          table: str | Table:

        Returns:
            Path to a CSV file in which to add values to a table on upload.
        """
        return table_path(
            self.working_dir,
            schema=self.domain_schema,
            table=self._get_table(table).name,
        )

    def asset_directory(self, table: str | Table, prefix: str | Path = None) -> Path:
        """Return a local file path in which to place a files for an asset table.  T

        Args:
            table: Location of where to place files.  Defaults to execution_assets_path.
            prefix: Root path to asset directory.

        Returns:
            Path to the directory in which asset files should be placed.
        """
        table = self._get_table(table)
        if not self.is_asset(table):
            raise DerivaMLException(f"The table {table} is not an asset table.")

        prefix = Path(prefix) or self.working_dir
        return asset_dir(prefix, table.schema.name, table.name)

    def download_dir(self, cached: bool = False) -> Path:
        """Location where downloaded files are placed.

        Args:
          cached: bool:  (Default value = False)

        Returns:

        """
        return self.cache_dir if cached else self.working_dir

    def chaise_url(self, table: RID | Table) -> str:
        """Return a Chaise URL to the specified table.

        Args:
            table: Table or RID to be visited
            table: str | Table:

        Returns:
            URL to the table in Chaise format.
        """
        try:
            table = self._get_table(table)
            uri = self.catalog.get_server_uri().replace(
                "ermrest/catalog/", "chaise/recordset/#"
            )
        except DerivaMLException:
            # Perhaps we have a RID....
            uri = self.cite(table)
        return f"{uri}/{urlquote(table.schema.name)}:{urlquote(table.name)}"

    def cite(self, entity: dict | str) -> str:
        """Return a citation URL for the provided entity.

        Args:
            entity: A dict that contains the column values for a specific entity or a RID.

        Returns:
            The URI for the provided entity.

        Raises:
            DerivaMLException: if provided RID does not exist.
        """
        if entity.startswith(f"https://{self.host_name}/id/{self.catalog_id}/"):
            # Already got a citation...
            return entity
        try:
            self.resolve_rid(
                rid := entity if isinstance(entity, str) else entity["RID"]
            )
            return f"https://{self.host_name}/id/{self.catalog_id}/{rid}@{self.catalog.latest_snapshot().snaptime}"
        except KeyError as e:
            raise DerivaMLException(f"Entity {e} does not have RID column")
        except DerivaMLException as _e:
            raise DerivaMLException("Entity RID does not exist")

    def user_list(self) -> list[dict[str, str]]:
        """List of users in the catalog

        Args:

        Returns:
          A list of dictionaries containing user information.

        """
        user_path = self.pathBuilder.public.ERMrest_Client.path
        return [
            {"ID": u["ID"], "Full_Name": u["Full_Name"]}
            for u in user_path.entities().fetch()
        ]

    def resolve_rid(self, rid: RID) -> ResolveRidResult:
        """Return a named tuple with information about the specified RID.

        Args:
            rid: RID of the object of interest

        Returns:
            ResolveRidResult which has information about the specified RID.

        Raises:
          DerivaMLException: if the RID doesn't exist.
        """
        try:
            return self.catalog.resolve_rid(rid, self.model)
        except KeyError as _e:
            raise DerivaMLException(f"Invalid RID {rid}")

    def retrieve_rid(self, rid: RID) -> dict[str, Any]:
        """Return a dictionary that represents the values of the specified RID.

        Args:
            rid: RID of the object of interest

        Returns:
          A dictionary that represents the values of the specified RID.

        Raises:
          DerivaMLException: if the RID doesn't exist.'
        """
        return self.resolve_rid(rid).datapath.entities().fetch()[0]

    def add_page(self, title: str, content: str) -> None:
        """

        Args:
          title: str:
          content: str:

        Returns:

        """
        self.pathBuilder.www.tables[self.domain_schema].insert(
            [{"Title": title, "Content": content}]
        )

    def create_vocabulary(
            self, vocab_name: str, comment: str = "", schema: Optional[str] = None
    ) -> Table:
        """Create a controlled vocabulary table with the given vocab name.

        Args:
            vocab_name: Name of the controlled vocabulary table.
            comment: Description of the vocabulary table. (Default value = '')
            schema: Schema in which to create the controlled vocabulary table.  Defaults to domain_schema.
            vocab_name: str:

        Returns:
            An ERMRest table object for the newly created vocabulary table.
        """
        schema = schema or self.domain_schema
        return self.model.schemas[schema].create_table(
            Table.define_vocabulary(
                vocab_name, f"{self.project_name}:{{RID}}", comment=comment
            )
        )

    def is_vocabulary(self, table_name: str | Table) -> bool:
        """Check if a given table is a controlled vocabulary table.

        Args:
          table_name: A ERMRest table object or the name of the table.
          table_name: str | Table:

        Returns:
          Table object if the table is a controlled vocabulary, False otherwise.

        Raises:
          DerivaMLException: if the table doesn't exist.

        """
        vocab_columns = {"NAME", "URI", "SYNONYMS", "DESCRIPTION", "ID"}
        table = self._get_table(table_name)
        return vocab_columns.issubset({c.name.upper() for c in table.columns})

    def create_asset(
            self, asset_name: str, comment: str = "", schema: str = None
    ) -> Table:
        """Create an asset table with the given asset name.

        Args:
            asset_name: Name of the asset table.
            comment: Description of the asset table. (Default value = '')
            schema: Schema in which to create the asset table.  Defaults to domain_schema.
            asset_name: str:
            schema: str:  (Default value = None)

        Returns:
            Table object for the asset table.
        """
        schema = schema or self.domain_schema
        asset_table = self.model.schemas[schema].create_table(
            Table.define_asset(schema, asset_name, comment=comment)
        )
        return asset_table

    def is_association(
            self, table_name: str | Table, unqualified: bool = True, pure: bool = True
    ) -> bool | set | int:
        """Check the specified table to see if it is an association table.

        Args:
            table_name: param unqualified:
            pure: return: (Default value = True)
            table_name: str | Table:
            unqualified:  (Default value = True)

        Returns:


        """
        table = self._get_table(table_name)
        return table.is_association(unqualified=unqualified, pure=pure)

    def is_asset(self, table_name: str | Table) -> bool:
        """True if the specified table is an asset table.

        Args:
            table_name: str | Table:

        Returns:
            True if the specified table is an asset table, False otherwise.

        """
        asset_columns = {"Filename", "URL", "Length", "MD5", "Description"}
        table = self._get_table(table_name)
        return asset_columns.issubset({c.name for c in table.columns})

    def find_assets(self) -> list[Table]:
        """ """
        return [
            t
            for s in self.model.schemas.values()
            for t in s.tables.values()
            if self.is_asset(t)
        ]

    @validate_call(config=ConfigDict(arbitrary_types_allowed=True))
    def create_feature(
            self,
            target_table: Table | str,
            feature_name: str,
            terms: list[Table | str] = None,
            assets: list[Table | str] = None,
            metadata: Iterable[ColumnDefinition | Table | Key | str] = None,
            optional: Optional[list[str]] = None,
            comment: str = "",
    ) -> type[FeatureRecord]:
        """Create a new feature that can be associated with a table.

        The feature can associate a controlled vocabulary term, an asset, or any other values with a s
        pecific instance of an object and  execution.

        Args:
            feature_name: Name of the new feature to be defined
            target_table: table name or object on which the feature is to be associated
            terms: List of controlled vocabulary terms that will be part of the feature value
            assets: List of asset table names or objects that will be part of the feature value
            metadata: List of other value types that are associated with the feature
            optional: List of columns that are optional in the feature
            comment: return: A Feature class that can be used to create instances of the feature.

        Returns:
            A Feature class that can be used to create instances of the feature.

        Raises:
            DerivaException: If the feature cannot be created.
        """

        terms = terms or []
        assets = assets or []
        metadata = metadata or []
        optional = optional or []

        def normalize_metadata(m: Key | Table | ColumnDefinition | str):
            """

            Args:
              m: Key | Table | ColumnDefinition | str:

            Returns:

            """
            if isinstance(m, str):
                return self._get_table(m)
            elif isinstance(m, ColumnDefinition):
                return m.model_dump()
            else:
                return m

        # Make sure that the provided assets or terms are actually assets or terms.
        if not all(map(self.is_asset, assets)):
            raise DerivaMLException(f"Invalid create_feature asset table.")
        if not all(map(self.is_vocabulary, terms)):
            raise DerivaMLException(f"Invalid create_feature asset table.")

        # Get references to the necessary tables and make sure that the
        # provided feature name exists.
        target_table = self._get_table(target_table)
        execution = self.model.schemas[self.ml_schema].tables["Execution"]
        feature_name_table = self.model.schemas[self.ml_schema].tables["Feature_Name"]
        feature_name_term = self.add_term(
            "Feature_Name", feature_name, description=comment
        )
        atable_name = f"Execution_{target_table.name}_{feature_name_term.name}"

        # Now create the association table that implements the feature.
        atable = self.model.schemas[self.domain_schema].create_table(
            target_table.define_association(
                table_name=atable_name,
                associates=[execution, target_table, feature_name_table],
                metadata=[
                    normalize_metadata(m) for m in chain(assets, terms, metadata)
                ],
                comment=comment,
            )
        )
        # Now set optional terms.
        for c in optional:
            atable.columns[c].alter(nullok=True)
        atable.columns["Feature_Name"].alter(default=feature_name_term.name)
        return self.feature_record_class(target_table, feature_name)

    def feature_record_class(
            self, table: str | Table, feature_name: str
    ) -> type[FeatureRecord]:
        """ "Create a pydantic model for entries into the specified feature table.

        For information on how to
        See the pydantic documentation for more details about the pydantic model.

        Args:
            table: table name or object on which the feature is to be associated
            feature_name: name of the feature to be created
            table: str | Table:
            feature_name: str:

        Returns:
            A Feature class that can be used to create instances of the feature.
        """
        return self.lookup_feature(table, feature_name).feature_record_class()

    def drop_feature(self, table: Table | str, feature_name: str) -> bool:
        """

        Args:
          table: Table | str:
          feature_name: str:

        Returns:
        """
        table = self._get_table(table)
        try:
            feature = next(
                f for f in self.find_features(table) if f.feature_name == feature_name
            )
            feature.feature_table.drop()
            return True
        except StopIteration:
            return False

    def lookup_feature(self, table: str | Table, feature_name: str) -> Feature:
        """Lookup the named feature associated with the provided table.

        Args:
            table: param feature_name:
            table: str | Table:
            feature_name: str:

        Returns:
            A Feature class that represents the requested feature.

        Raises:
          DerivaMLException: If the feature cannot be found.
        """
        table = self._get_table(table)
        try:
            return [
                f for f in self.find_features(table) if f.feature_name == feature_name
            ][0]
        except IndexError:
            raise DerivaMLException(
                f"Feature {table.name}:{feature_name} doesn't exist."
            )

    @validate_call(config=ConfigDict(arbitrary_types_allowed=True))
    def find_features(self, table: Table | str) -> Iterable[Feature]:
        """List the names of the features in the specified table.

        Args:
            table: The table to find features for.
            table: Table | str:

        Returns:
            An iterable of FeatureResult instances that describe the current features in the table.
        """
        table = self._get_table(table)

        def is_feature(a: FindAssociationResult) -> bool:
            """

            Args:
              a: FindAssociationResult:

            Returns:

            """
            # return {'Feature_Name', 'Execution'}.issubset({c.name for c in a.table.columns})
            return {
                "Feature_Name",
                "Execution",
                a.self_fkey.foreign_key_columns[0].name,
            }.issubset({c.name for c in a.table.columns})

        return [
            Feature(a)
            for a in table.find_associations(min_arity=3, max_arity=3, pure=False)
            if is_feature(a)
        ]

    @validate_call
    def add_features(self, features: Iterable[FeatureRecord]) -> int:
        """Add a set of new feature values to the catalog.

        Args:
          features: Iterable[FeatureRecord]:

        Returns:
            Number of attributes added
        """
        features = list(features)
        feature_table = features[0].feature.feature_table
        feature_path = self.pathBuilder.schemas[feature_table.schema.name].tables[
            feature_table.name
        ]
        entries = feature_path.insert(f.model_dump() for f in features)
        return len(entries)

    @validate_call(config=ConfigDict(arbitrary_types_allowed=True))
    def list_feature_values(self, table: Table | str, feature_name: str) -> _ResultSet:
        """Return a datapath resultset containing all values of a feature associated with a table.

        Args:
            table: param feature_name:
            table: Table | str:
            feature_name: str:

        Returns:

        """
        table = self._get_table(table)
        feature = self.lookup_feature(table, feature_name)
        pb = self.catalog.getPathBuilder()
        return (
            pb.schemas[feature.feature_table.schema.name]
            .tables[feature.feature_table.name]
            .entities()
            .fetch()
        )

    @validate_call
    def create_dataset(
            self,
            ds_type: str | list[str],
            description: str,
            execution_rid: Optional[RID] = None,
            version: tuple[int, int, int] = (1, 0, 0),
    ) -> RID:
        """Create a new dataset from the specified list of RIDs.

        Args:
            ds_type: One or more dataset types.  Must be a term from the DatasetType controlled vocabulary.
            description: Description of the dataset.
            execution_rid: Execution under which the dataset will be created.
            version: Version of the dataset.
            ds_type: str | list[str]:
            description: str:
            execution_rid: Optional[RID]:  (Default value = None)
            version: tuple[int: int: int]

        Returns:
            New dataset RID.

        """
        # Create the entry for the new dataset and get its RID.
        ds_types = [ds_type] if isinstance(ds_type, str) else ds_type

        pb = self.pathBuilder
        for ds_type in ds_types:
            if not self.lookup_term(MLVocab.dataset_type, ds_type):
                raise DerivaMLException(f"Dataset type must be a vocabulary term.")
        dataset_table_path = pb.schemas[self.dataset_table.schema.name].tables[
            self.dataset_table.name
        ]
        dataset = dataset_table_path.insert(
            [
                {
                    "Description": description,
                    MLVocab.dataset_type: ds_type,
                    "Version": f"{version[0]}.{version[1]}.{version[2]}",
                }
            ]
        )[0]["RID"]

        # Get the name of the association table between dataset and dataset_type.
        atable = next(
            self.model.schemas[self.ml_schema]
            .tables[MLVocab.dataset_type]
            .find_associations()
        ).name
        pb.schemas[self.ml_schema].tables[atable].insert(
            [
                {MLVocab.dataset_type: ds_type, "Dataset": dataset}
                for ds_type in ds_types
            ]
        )
        if execution_rid is not None:
            pb.schemas[self.ml_schema].Dataset_Execution.insert(
                [{"Dataset": dataset, "Execution": execution_rid}]
            )
        return dataset

    def find_datasets(self) -> Iterable[dict[str, Any]]:
        """Returns a list of currently available datasets.

        Returns:
             list of currently available datasets.
        """
        # Get datapath to all the tables we will need: Dataset, DatasetType and the association table.
        pb = self.pathBuilder
        dataset_path = pb.schemas[self.dataset_table.schema.name].tables[
            self.dataset_table.name
        ]
        atable = next(
            self.model.schemas[self.ml_schema]
            .tables[MLVocab.dataset_type]
            .find_associations()
        ).name
        ml_path = pb.schemas[self.ml_schema]
        atable_path = ml_path.tables[atable]

        # Get a list of all the dataset_type values associated with this dataset.
        datasets = []
        for dataset in dataset_path.entities().fetch():
            ds_types = (
                atable_path.filter(atable_path.Dataset == dataset["RID"])
                .attributes(atable_path.Dataset_Type)
                .fetch()
            )
            datasets.append(
                dataset
                | {MLVocab.dataset_type: [ds[MLVocab.dataset_type] for ds in ds_types]}
            )
        return datasets

    @validate_call
    def delete_dataset(self, dataset_rid: RID, recurse: bool = False) -> None:
        """Delete a dataset from the catalog.

        Args:
            dataset_rid: RID of the dataset to delete.
            recurse: If True, delete the dataset along with any nested datasets. (Default value = False)
            dataset_rid: RID:
        """
        # Get association table entries for this dataset
        # Delete association table entries
        pb = self.pathBuilder
        for assoc_table in self.dataset_table.find_associations(self.dataset_table):
            other_fkey = assoc_table.other_fkeys.pop()
            self_fkey = assoc_table.self_fkey
            target_table = other_fkey.pk_table
            member_table = assoc_table.table

            schema_path = pb.schemas[member_table.schema.name]
            tpath = schema_path.tables[assoc_table.name]
            dataset_column_path = tpath.columns[self_fkey.columns[0].name]
            dataset_entries = tpath.filter(dataset_column_path == dataset_rid)
            if recurse and target_table == self.dataset_table:
                # Nested table
                for dataset in dataset_entries:
                    self.delete_dataset(dataset["RID"], recurse)
            try:
                dataset_entries.delete()
            except DataPathException:
                pass

        # Delete dataset.
        dataset_path = pb.schemas[self.dataset_table.schema.name].tables[
            self.dataset_table.name
        ]
        dataset_path.filter(dataset_path.columns["RID"] == dataset_rid).delete()

    def list_dataset_element_types(self) -> Iterable[Table]:
        """List the types of entities that can be added to a dataset.

        Returns:
          :return: An iterable of Table objects that can be included as an element of a dataset.
        """

        def domain_table(table: Table) -> bool:
            return (
                    table.schema.name == self.domain_schema
                    or table.name == self.dataset_table.name
            )

        return [
            t
            for a in self.dataset_table.find_associations()
            if domain_table(t := a.other_fkeys.pop().pk_table)
        ]

    @validate_call(config=ConfigDict(arbitrary_types_allowed=True))
    def add_dataset_element_type(self, element: str | Table) -> Table:
        """A dataset is a heterogeneous collection of objects, each of which comes from a different table. This
        routine makes it possible to add objects from the specified table to a dataset.

        Args:
            element: Name or the table or table object that is to be added to the dataset.
            element: str | Table:

        Returns:
            The table object that was added to the dataset.
        """
        # Add table to map
        element_table = self._get_table(element)
        table = self.model.schemas[self.domain_schema].create_table(
            Table.define_association([self.dataset_table, element_table])
        )

        # self.model = self.catalog.getCatalogModel()
        self.dataset_table.annotations.update(generate_dataset_annotations(self.model))
        self.model.apply()
        return table

    @validate_call
    def list_dataset_parents(self, dataset_rid: RID) -> list[RID]:
        """Given a dataset RID, return a list of RIDs of the parent datasets.

        Args:
            dataset_rid: return: RID of the parent dataset.
            dataset_rid: RID:

        Returns:
            RID of the parent dataset.
        """
        rid_record = self.resolve_rid(dataset_rid)
        if rid_record.table.name != self.dataset_table.name:
            raise DerivaMLException(
                f"RID: {dataset_rid} does not belong to dataset {self.dataset_table.name}"
            )
        # Get association table for nested datasets
        atable_path = self.pathBuilder.schemas[self.ml_schema].Dataset_Dataset
        return [
            p["Dataset"]
            for p in atable_path.filter(atable_path.Nested_Dataset == dataset_rid)
            .entities()
            .fetch()
        ]

    @validate_call
    def list_dataset_children(self, dataset_rid: RID) -> list[RID]:
        """Given a dataset RID, return a list of RIDs of any nested datasets.

        Args:
            dataset_rid: A dataset RID.

        Returns:
          list of RIDs of nested datasets.

        """
        return [d["RID"] for d in self.list_dataset_members(dataset_rid)["Dataset"]]

    @validate_call
    def list_dataset_members(
            self, dataset_rid: RID, recurse: bool = False
    ) -> dict[str, list[dict[str, Any]]]:
        """Return a list of entities associated with a specific dataset.

        Args:
            dataset_rid: param recurse: If this is a nested dataset, list the members of the contained datasets
            dataset_rid: RID:
            recurse:  (Default value = False)

        Returns:
            Dictionary of entities associated with a specific dataset.  Key is the table from which the elements
            were taken.
        """

        try:
            if self.resolve_rid(dataset_rid).table != self.dataset_table:
                raise DerivaMLException(f"RID is not for a dataset: {dataset_rid}")
        except DerivaMLException:
            raise DerivaMLException(f"Invalid RID: {dataset_rid}")

        # Look at each of the element types that might be in the dataset and get the list of rid for them from
        # the appropriate association table.
        members = defaultdict(list)
        pb = self.pathBuilder
        for assoc_table in self.dataset_table.find_associations():
            other_fkey = assoc_table.other_fkeys.pop()
            self_fkey = assoc_table.self_fkey
            target_table = other_fkey.pk_table
            member_table = assoc_table.table

            if (
                    target_table.schema.name != self.domain_schema
                    and target_table != self.dataset_table
            ):
                # Look at domain tables and nested datasets.
                continue
            if target_table == self.dataset_table:
                # find_assoc gives us the keys in the wrong position, so swap.
                self_fkey, other_fkey = other_fkey, self_fkey

            target_path = pb.schemas[target_table.schema.name].tables[target_table.name]
            member_path = pb.schemas[member_table.schema.name].tables[member_table.name]
            # Get the names of the columns that we are going to need for linking
            member_link = tuple(
                c.name for c in next(iter(other_fkey.column_map.items()))
            )
            path = pb.schemas[member_table.schema.name].tables[member_table.name].path
            path.filter(member_path.Dataset == dataset_rid)
            path.link(
                target_path,
                on=(
                        member_path.columns[member_link[0]]
                        == target_path.columns[member_link[1]]
                ),
            )
            target_entities = path.entities().fetch()
            members[target_table.name].extend(target_entities)
            if recurse and target_table == self.dataset_table:
                # Get the members for all the nested datasets and add to the member list.
                nested_datasets = [d["RID"] for d in target_entities]
                for ds in nested_datasets:
                    for k, v in self.list_dataset_members(ds, recurse=False).items():
                        members[k].extend(v)
        return members

    @validate_call
    def add_dataset_members(
            self, dataset_rid: Optional[RID], members: list[RID], validate: bool = True
    ) -> None:
        """Add additional elements to an existing dataset.

        Args:
            dataset_rid: RID of dataset to extend or None if new dataset is to be created.
            members: List of RIDs of members to add to the  dataset.
            validate: Check rid_list to make sure elements are not already in the dataset.
            dataset_rid: Optional[RID]:
            members: list[RID]:
            validate: bool:  (Default value = True)
        """

        members = set(members)

        def check_dataset_cycle(member_rid, path=None):
            """

            Args:
              member_rid:
              path:  (Default value = None)

            Returns:

            """
            path = path or set(dataset_rid)
            return member_rid in path

        if validate:
            existing_rids = set(
                m["RID"]
                for ms in self.list_dataset_members(dataset_rid).values()
                for m in ms
            )
            if overlap := set(existing_rids).intersection(members):
                raise DerivaMLException(
                    f"Attempting to add existing member to dataset {dataset_rid}: {overlap}"
                )

        # Now go through every rid to be added to the data set and sort them based on what association table entries
        # need to be made.
        dataset_elements = {}
        association_map = {
            a.other_fkeys.pop().pk_table.name: a.table.name
            for a in self.dataset_table.find_associations()
        }
        # Get a list of all the types of objects that can be linked to a dataset.
        for m in members:
            rid_info = self.resolve_rid(m)
            if rid_info.table.name not in association_map:
                raise DerivaMLException(
                    f"RID table: {rid_info.table.name} not part of dataset"
                )
            if rid_info.table == self.dataset_table and check_dataset_cycle(
                    rid_info.rid
            ):
                raise DerivaMLException("Creating cycle of datasets is not allowed")
            dataset_elements.setdefault(rid_info.table.name, []).append(rid_info.rid)
        # Now make the entries into the association tables.
        pb = self.pathBuilder
        for table, elements in dataset_elements.items():
            schema_path = pb.schemas[
                self.ml_schema if table == "Dataset" else self.domain_schema
            ]
            fk_column = "Nested_Dataset" if table == "Dataset" else table

            if len(elements):
                # Find out the name of the column in the association table.
                schema_path.tables[association_map[table]].insert(
                    [{"Dataset": dataset_rid, fk_column: e} for e in elements]
                )

    @validate_call(config=ConfigDict(arbitrary_types_allowed=True))
    def add_term(
            self,
            table: str | Table,
            term_name: str,
            description: str,
            synonyms: Optional[list[str]] = None,
            exists_ok: bool = True,
    ) -> VocabularyTerm:
        """Creates a new control vocabulary term in the control vocabulary table.

        Args:

        Args:
            table: The name of the control vocabulary table.
            term_name: The name of the new control vocabulary.
            description: The description of the new control vocabulary.
            synonyms: Optional list of synonyms for the new control vocabulary. Defaults to an empty list.
            exists_ok: Optional flag indicating whether to allow creation if the control vocabulary name
                already exists. Defaults to True.

        Returns:
          The RID of the newly created control vocabulary.

        Raises:
          DerivaException: If the control vocabulary name already exists and exist_ok is False.
        """
        synonyms = synonyms or []
        table = self._get_table(table)
        pb = self.catalog.getPathBuilder()
        if not (self.is_vocabulary(table)):
            raise DerivaMLException(f"The table {table} is not a controlled vocabulary")

        schema_name = table.schema.name
        table_name = table.name
        try:
            term_id = VocabularyTerm.model_validate(
                pb.schemas[schema_name]
                .tables[table_name]
                .insert(
                    [
                        {
                            "Name": term_name,
                            "Description": description,
                            "Synonyms": synonyms,
                        }
                    ],
                    defaults={"ID", "URI"},
                )[0]
            )
        except DataPathException:
            term_id = self.lookup_term(table, term_name)
            if not exists_ok:
                raise DerivaMLException(f"{term_name} already exists")
            # Check vocabulary
        return term_id

    @validate_call(config=ConfigDict(arbitrary_types_allowed=True))
    def lookup_term(self, table: str | Table, term_name: str) -> VocabularyTerm:
        """Given a term name, return the vocabulary record.  Can provide either the term name
         or a synonym for the term.  Generate an exception if the term is not in the vocabulary.

        Args:
            table: The name of the controlled vocabulary table or a ERMrest table object..
            term_name: The name of the term to look up.

        Returns:
          The entry the associated term or synonym.

        Raises:
          DerivaException: If the schema or vocabulary table doesn't exist, or if the term is not
            found in the vocabulary.
        """
        vocab_table = self._get_table(table)
        if not self.is_vocabulary(vocab_table):
            raise DerivaMLException(f"The table {table} is not a controlled vocabulary")
        schema_name, table_name = vocab_table.schema.name, vocab_table.name
        schema_path = self.catalog.getPathBuilder().schemas[schema_name]
        for term in schema_path.tables[table_name].entities():
            if term_name == term["Name"] or (
                    term["Synonyms"] and term_name in term["Synonyms"]
            ):
                return VocabularyTerm.model_validate(term)
        raise DerivaMLException(f"Term {term_name} is not in vocabulary {table_name}")

    def find_vocabularies(self) -> Iterable[Table]:
        """Return a list of all the controlled vocabulary tables in the domain schema."""
        return [
            t
            for s in self.model.schemas.values()
            for t in s.tables.values()
            if self.is_vocabulary(t)
        ]

    def list_vocabulary_terms(self, table: str | Table) -> list[VocabularyTerm]:
        """Return an list of terms that are in a vocabulary table.

        Args:
            table: The name of the controlled vocabulary table or a ERMrest table object.
            table: str | Table:

        Returns:
            The list of terms that are in a vocabulary table.

        Raises:
            DerivaMLException: If the schema or vocabulary table doesn't exist, or if the table is not
                a controlled vocabulary.
        """
        pb = self.catalog.getPathBuilder()
        table = self._get_table(table)
        if not (self.is_vocabulary(table)):
            raise DerivaMLException(f"The table {table} is not a controlled vocabulary")

        return [
            VocabularyTerm(**v)
            for v in pb.schemas[table.schema.name].tables[table.name].entities().fetch()
        ]

    @validate_call
    def download_dataset_bag(
            self,
            bag: RID | str,
            materialize: bool = True,
            execution_rid: Optional[RID] = None,
    ) -> tuple[Path, RID]:
        """Given a RID to a dataset, or a MINID to an existing bag, download the bag file, extract it and validate
        that all the metadata is correct

        Args:
            bag: The RID of a dataset or a minid to an existing bag.
            materialize: Materalize the bag, rather than just downloading it.
            execution_rid: RID of execution object requesting the download.  Used to update status.

        Returns:
            the location of the unpacked and validated dataset bag and the RID of the bag
        """
        return (
            self._materialize_dataset_bag(bag, execution_rid)
            if materialize
            else self._download_dataset_bag(bag)
        )

    def _download_dataset_bag(self, dataset_rid: RID | str) -> tuple[Path, RID]:
        """Given a RID to a dataset, or a MINID to an existing bag, download the bag file, extract it and validate
        that all the metadata is correct

        Args:
            dataset_rid: The RID of a dataset or a minid to an existing bag.
             dataset_rid: RID | str:

        Returns:
            the location of the unpacked and validated dataset bag and the RID of the bag
        """
        if not any([dataset_rid == ds["RID"] for ds in self.find_datasets()]):
            raise DerivaMLException(f"RID {dataset_rid} is not a dataset")

        with TemporaryDirectory() as tmp_dir:
            if dataset_rid.startswith("minid"):
                # If provided a MINID, use the MINID metadata to get the checksum and download the bag.
                r = requests.get(
                    f"https://identifiers.org/{dataset_rid}",
                    headers={"accept": "application/json"},
                )
                metadata = r.json()["metadata"]
                dataset_rid = metadata["Dataset_RID"].split("@")[0]
                checksum_value = ""
                for checksum in r.json().get("checksums", []):
                    if checksum.get("function") == "sha256":
                        checksum_value = checksum.get("value")
                        break
                archive_path = fetch_single_file(dataset_rid, tmp_dir)
            else:
                # We are given the RID to a dataset, so we are going to have to export as a bag and place into
                # local file system.  The first step is to generate a downloadspec to create the bag, put the sped
                # into a local file and then use the downloader to create and download the desired bdbag.
                spec_file = f"{tmp_dir}/download_spec.json"
                with open(spec_file, "w", encoding="utf-8") as ds:
                    json.dump(Dataset(self.model).generate_dataset_download_spec(), ds)
                downloader = GenericDownloader(
                    server={
                        "catalog_id": self.catalog_id,
                        "protocol": "https",
                        "host": self.host_name,
                    },
                    config_file=spec_file,
                    output_dir=tmp_dir,
                    envars={"Dataset_RID": dataset_rid},
                )
                result = downloader.download()
                archive_path = list(result.values())[0]["local_path"]
                checksum_value = compute_file_hashes(archive_path, hashes=["sha256"])[
                    "sha256"
                ][0]

            # Check to see if we have an existing idempotent materialization of the desired bag. If so, then just reuse
            # it.  If not, then we need to extract the contents of the archive into our cache directory.
            bag_dir = self.cache_dir / f"{dataset_rid}_{checksum_value}"
            if bag_dir.exists():
                bag_path = (bag_dir / f"Dataset_{dataset_rid}").as_posix()
            else:
                bag_dir.mkdir(parents=True, exist_ok=True)
                bag_path = bdb.extract_bag(archive_path, bag_dir)
            bdb.validate_bag_structure(bag_path)
            return Path(bag_path), dataset_rid

    def _materialize_dataset_bag(
            self, bag: str | RID, execution_rid: Optional[RID] = None
    ) -> tuple[Path, RID]:
        """Materialize a dataset bag into a local directory

        Args:
            bag: A MINID to an existing bag or a RID of the dataset that should be downloaded.
            execution_rid: RID of the execution for which this bag should be materialized. Used to update status.
            bag: str | RID:
            execution_rid: Optional[RID]:  (Default value = None)

        Returns:

        """

        def fetch_progress_callback(current, total):
            """

            Args:
              current:
              total:

            Returns:

            """
            msg = f"Materializing bag: {current} of {total} file(s) downloaded."
            if execution_rid:
                self._update_status(Status.running, msg, execution_rid)
            logging.info(msg)
            return True

        def validation_progress_callback(current, total):
            """

            Args:
              current:
              total:

            Returns:

            """
            msg = f"Validating bag: {current} of {total} file(s) validated."
            if execution_rid:
                self._update_status(Status.running, msg, execution_rid)
            logging.info(msg)
            return True

        if execution_rid and self.resolve_rid(execution_rid).table.name != "Execution":
            raise DerivaMLException(f"RID {execution_rid} is not an execution")

        # request metadata
        bag_path, dataset_rid = self._download_dataset_bag(bag)
        bag_dir = bag_path.parent
        validated_check = bag_dir / "validated_check.txt"

        # If this bag has already been validated, our work is done.  Otherwise, materialize the bag.
        if not validated_check.exists():
            bdb.materialize(
                bag_path.as_posix(),
                fetch_callback=fetch_progress_callback,
                validation_callback=validation_progress_callback,
            )
            validated_check.touch()
        return Path(bag_path), dataset_rid

    def download_asset(self, asset_url: str, dest_filename: str) -> Path:
        """Download an asset from a URL and place it in a local directory.

        Args:
            asset_url: URL of the asset.
            dest_filename: Destination filename.

        Returns:
            A  Path object to the downloaded asset.
        """
        hs = HatracStore("https", self.host_name, self.credential)
        hs.get_obj(path=asset_url, destfilename=dest_filename)
        return Path(dest_filename)

    @validate_call(config=ConfigDict(arbitrary_types_allowed=True))
    def upload_asset(self, file: Path | str, table: Table | str, **kwargs: Any) -> dict:
        """Upload the specified file into Hatrac and update the associated asset table.

        Args:
            file: path to the file to upload.
            table: Name of the asset table
            kwargs: Keyword arguments for values of additional columns to be added to the asset table.

        Returns:

        """
        table = self._get_table(table)
        if not self.is_asset(table):
            raise DerivaMLException(f"Table {table} is not an asset table.")

        credential = self.model.catalog.deriva_server.credentials
        file_path = Path(file)
        file_name = file_path.name
        file_size = file_path.stat().st_size
        # Get everything up to the filename  part of the
        hatrac_path = f"/hatrac/{table.name}/"
        hs = HatracStore("https", self.host_name, credential)
        md5_hashes = hash_utils.compute_file_hashes(file, ["md5"])["md5"]
        sanitized_filename = urlquote(
            re.sub("[^a-zA-Z0-9_.-]", "_", md5_hashes[0] + "." + file_name)
        )
        hatrac_path = f"{hatrac_path}{sanitized_filename}"

        try:
            # Upload the file to hatrac.
            hatrac_uri = hs.put_obj(
                hatrac_path,
                file,
                md5=md5_hashes[1],
                content_type=mime_utils.guess_content_type(file),
                content_disposition="filename*=UTF-8''" + file_name,
            )
        except Exception as e:
            raise e
        try:
            # Now update the asset table.
            ipath = self.pathBuilder.schemas[table.schema.name].tables[table.name]
            return list(
                ipath.insert(
                    [
                        {
                            "URL": hatrac_uri,
                            "Filename": file_name,
                            "Length": file_size,
                            "MD5": md5_hashes[0],
                        }
                        | kwargs
                    ]
                )
            )[0]
        except Exception as e:
            raise e

    def upload_assets(self, assets_dir: str | Path) -> dict[str, FileUploadState]:
        """Upload assets from a directory. This routine assumes that the current upload specification includes a
        configuration for the specified directory.  Every asset in the specified directory is uploaded

        Args:
            assets_dir: Directory containing the assets to upload.

        Returns:
            Results of the upload operation.

        Raises:
            DerivaMLException: If there is an issue uploading the assets.
        """
        with TemporaryDirectory() as temp_dir:
            spec_file = f"{temp_dir}/config.json"
            with open(spec_file, "w+") as cfile:
                json.dump(bulk_upload_configuration, cfile)
            uploader = GenericUploader(
                server={
                    "host": self.host_name,
                    "protocol": "https",
                    "catalog_id": self.catalog_id,
                },
                config_file=spec_file,
            )
            uploader.getUpdatedConfig()
            uploader.scanDirectory(assets_dir)
            results = {
                path: FileUploadState(
                    state=UploadState(result["State"]),
                    status=result["Status"],
                    result=result["Result"],
                )
                for path, result in uploader.uploadFiles().items()
            }
            uploader.cleanup()
        return results

    def _update_status(
            self, new_status: Status, status_detail: str, execution_rid: RID
    ):
        """Update the status of an execution in the catalog.

        Args:
            new_status: New status.
            status_detail: Details of the status.
            execution_rid: Resource Identifier (RID) of the execution.
            new_status: Status:
            status_detail: str:
             execution_rid: RID:

        Returns:

        """
        self.status = new_status.value
        self.pathBuilder.schemas[self.ml_schema].Execution.update(
            [
                {
                    "RID": execution_rid,
                    "Status": self.status,
                    "Status_Detail": status_detail,
                }
            ]
        )

    def upload_execution_configuration(self, config: ExecutionConfiguration) -> RID:
        """Upload execution configuration to Execution_Metadata table with Execution Metadata Type = Execution_Config.

        Args:
            config: A execution configuration.

        Returns:
            RID of the newly created configuration file.

        Raises:
             DerivaMLException: If there is an issue uploading the configuration.
        """
        try:
            fp = NamedTemporaryFile(
                "w+", prefix="exec_config", suffix=".json", delete=False
            )
            json.dump(config.model_dump_json(), fp)
            fp.close()
            configuration_rid = self._upload_execution_configuration_file(
                fp.name, description=config.description
            )
            os.remove(fp.name)
        except Exception as _e:
            raise DerivaMLException(f"Error in execution configuration upload")
        return configuration_rid

    def download_execution_configuration(
            self, configuration_rid: RID
    ) -> ExecutionConfiguration:
        """Create an ExecutionConfiguration object from a catalog RID that points to a JSON representation of that
        configuration in hatrac

        Args:
            configuration_rid: RID that should be to an asset table that refers to an execution configuration

        Returns:
            A ExecutionConfiguration object for configured by the parameters in the configuration file.
        """
        configuration = self.retrieve_rid(configuration_rid)
        with NamedTemporaryFile("w+", delete=False, suffix=".json") as dest_file:
            hs = HatracStore("https", self.host_name, self.credential)
            hs.get_obj(path=configuration["URL"], destfilename=dest_file.name)
            return ExecutionConfiguration.load_configuration(dest_file.name)

    def _upload_execution_configuration_file(
            self, config_file: str, description: str
    ) -> RID:
        """

        Args:
            config_file: str:
            description: str:

        Returns:

        """
        file_path = Path(config_file)
        file_name = file_path.name
        file_size = file_path.stat().st_size
        try:
            hs = HatracStore("https", self.host_name, self.credential)
            md5 = hash_utils.compute_file_hashes(config_file, ["md5"])["md5"][1]
            sanitized_filename = urlquote(
                re.sub("[^a-zA-Z0-9_.-]", "_", md5 + "." + file_name)
            )
            hatrac_path = f"/hatrac/execution_metadata/{sanitized_filename}"
            hatrac_uri = hs.put_obj(
                hatrac_path,
                config_file,
                md5=md5,
                content_type=mime_utils.guess_content_type(config_file),
                content_disposition="filename*=UTF-8''" + file_name,
            )
        except Exception as e:
            error = format_exception(e)
            raise DerivaMLException(
                f"Failed to upload execution configuration file {config_file} to object store. Error: {error}"
            )
        try:
            ml_schema_path = self.pathBuilder.schemas[self.ml_schema]
            return list(
                ml_schema_path.tables["Execution_Metadata"].insert(
                    [
                        {
                            "URL": hatrac_uri,
                            "Filename": file_name,
                            "Length": file_size,
                            "MD5": md5,
                            "Description": description,
                            "Execution_Metadata_Type": ExecMetadataVocab.execution_config,
                        }
                    ]
                )
            )[0]["RID"]
        except Exception as e:
            error = format_exception(e)
            raise DerivaMLException(
                f"Failed to update Execution_Asset table with configuration file metadata. Error: {error}"
            )

    # @validate_call
    def create_execution(self, configuration: ExecutionConfiguration) -> "Execution":
        """Create an execution object

        Args:
            configuration: ExecutionConfiguration:

        Returns:
            An execution object.
        """
        from .execution import Execution

        return Execution(configuration, self)



"""
Shared definitions that are used in different DerivaML modules.
"""


import warnings
from enum import Enum
from typing import Any, Iterable, Optional, Annotated

import deriva.core.ermrest_model as em
from deriva.core.ermrest_model import builtin_types
from pydantic import BaseModel, model_serializer, Field, computed_field

ML_SCHEMA = 'deriva-ml'

# We are going to use schema as a field name and this collides with method in pydantic base class
warnings.filterwarnings('ignore',
                        message='Field name "schema"',
                        category=Warning,
                        module='pydantic')

warnings.filterwarnings('ignore',
                        message='fields may not start with an underscore',
                        category=Warning,
                        module='pydantic')


rid_regex = r"^(?:[A-Z\d]{1,4}|[A-Z\d]{1,4}(?:-[A-Z\d]{4})+)$"
RID = Annotated[str, Field(pattern=rid_regex)]

DerivaSystemColumns = ['RCT', 'RMT', 'RCB', 'RMB']

# For some reason, deriva-py doesn't use the proper enum class!!
class UploadState(Enum):
    success = 0
    failed = 1
    pending = 2
    running = 3
    paused = 4
    aborted = 5
    cancelled = 6
    timeout = 7

class StrEnum(str, Enum):
    pass

class FileUploadState(BaseModel):
    state: UploadState
    status: str
    result: Any

    @computed_field
    @property
    def rid(self) -> Optional[RID]:
        return self.result and self.result['RID']

class Status(StrEnum):
    """Enumeration class defining execution status.

    Attributes:
        running: Execution is currently running.
        pending: Execution is pending.
        completed: Execution has been completed successfully.
        failed: Execution has failed.

    """
    running = 'Running'
    pending = 'Pending'
    completed = 'Completed'
    failed = 'Failed'


class BuiltinTypes(Enum):
    text = builtin_types.text
    int2 = builtin_types.int2
    jsonb = builtin_types.json
    float8 = builtin_types.float8
    timestamp = builtin_types.timestamp
    int8 = builtin_types.int8
    boolean = builtin_types.boolean
    json = builtin_types.json
    float4 = builtin_types.float4
    int4 = builtin_types.int4
    timestamptz = builtin_types.timestamptz
    date = builtin_types.date
    ermrest_rid = builtin_types.ermrest_rid
    ermrest_rcb = builtin_types.ermrest_rcb
    ermrest_rmb = builtin_types.ermrest_rmb
    ermrest_rct = builtin_types.ermrest_rct
    ermrest_rmt = builtin_types.ermrest_rmt
    markdown = builtin_types.markdown
    longtext = builtin_types.longtext
    ermrest_curie = builtin_types.ermrest_curie
    ermrest_uri = builtin_types.ermrest_uri
    color_rgb_hex = builtin_types.color_rgb_hex
    serial2 = builtin_types.serial2
    serial4 = builtin_types.serial4
    serial8 = builtin_types.serial8

class MLVocab(StrEnum):
    """Names of controlled vocabulary for various types within DerivaML."""
    dataset_type = 'Dataset_Type'
    workflow_type = 'Workflow_Type'
    execution_asset_type = 'Execution_Asset_Type'
    execution_metadata_type = 'Execution_Metadata_Type'

class ExecMetadataVocab(StrEnum):
    """
    Predefined execution metatadata types.
    """
    execution_config = 'Execution_Config'
    runtime_env = 'Runtime_Env'

class ColumnDefinition(BaseModel):
    """Pydantic model for deriva_py Column.define"""

    name: str
    type: BuiltinTypes
    nullok: bool = True
    default: Any = None
    comment: str = None
    acls: dict = Field(default_factory=dict)
    acl_bindings: dict = Field(default_factory=dict)
    annotations: dict = Field(default_factory=dict)

    @model_serializer()
    def serialize_column_definition(self):
        return em.Column.define(
            cname=self.name,
            ctype=self.type.value,
            nullok=self.nullok,
            default=self.default,
            comment=self.comment,
            acls=self.acls,
            acl_bindings=self.acl_bindings,
            annotations=self.annotations)


class KeyDefinition(BaseModel):
    colnames: Iterable[str]
    constraint_names: Iterable[str]
    comment: Optional[str] = None
    annotations: dict = Field(default_factory=dict)

    @model_serializer()
    def serialize_key_definition(self):
        return em.Key.define(
            colnames=self.colnames,
            constraint_names=self.constraint_names,
            comment=self.comment,
            annotations=self.annotations
        )


class ForeignKeyDefinition(BaseModel):
    """Pydantic model for deriva_py ForeignKey.define"""
    colnames: Iterable[str]
    pk_sname: str
    pk_tname: str
    pk_colnames: Iterable[str]
    constraint_names: Iterable[str] = Field(default_factory=list)
    on_update: str = 'NO ACTION'
    on_delete: str = 'NO ACTION'
    comment: str = None
    acls: dict[str, Any] = Field(default_factory=dict)
    acl_bindings: dict[str, Any] = Field(default_factory=dict)
    annotations: dict[str, Any] = Field(default_factory=dict)

    @model_serializer()
    def serialize_fk_definition(self):
        return em.ForeignKey.define(
            fk_colnames=self.colnames,
            pk_sname=self.pk_sname,
            pk_tname=self.pk_tname,
            pk_colnames=self.pk_colnames,
            on_update=self.on_update,
            on_delete=self.on_delete,
            comment=self.comment,
            acls=self.acls,
            acl_bindings=self.acl_bindings,
            annotations=self.annotations
        )


class TableDefinition(BaseModel):
    name: str
    column_defs: Iterable[ColumnDefinition]
    key_defs: Iterable[KeyDefinition] = Field(default_factory=list)
    fkey_defs: Iterable[ForeignKeyDefinition] = Field(default_factory=list)
    comment: str = None
    acls: dict = Field(default_factory=dict)
    acl_bindings: dict = Field(default_factory=dict)
    annotations: dict = Field(default_factory=dict)

    @model_serializer()
    def serialize_table_definition(self):
        return em.Table.define(
            tname=self.name,
            column_defs=[c.model_dump() for c in self.column_defs],
            key_defs=[k.model_dump() for k in self.key_defs],
            fkey_defs=[fk.model_dump() for fk in self.fkey_defs],
            comment=self.comment,
            acls=self.acls,
            acl_bindings=self.acl_bindings,
            annotations=self.annotations)


class DerivaMLException(Exception):
    """Exception class specific to DerivaML module.

    Args:
        msg (str): Optional message for the exception.
    """

    def __init__(self, msg=''):
        super().__init__(msg)
        self._msg = msg



"""
This module provides functions that help stucture local directories for uploading to a DerivaML catalog, and
generating an upload specification for those directories.

Here is the directory layout we support:

  deriva-ml/
       execution
           <execution_rid>
               execution-asset
                   <asset_type>
                       file1, file2, ....   <- Need to update execution_asset association table.
               execution-metadata
                   <metadata_type>
               feature
                   <schema>
                       <target_table>
                            <feature_name>
                                   asset
                                       <asset_table>
                                           file1, file2, ...
                           <feature_name>.csv    <- needs to have asset_name column remapped before uploading
            table
               <schema>
                   <record_table>
                      record_table.csv
            asset
               <schema>
                   <asset_table>
                     file1, file2, ....

"""

from pathlib import Path
from typing import Optional

import regex as re

upload_root_regex = r"(?i)^.*/deriva-ml"
exec_dir_regex = upload_root_regex + r"/execution/(?P<execution_rid>[-\w]+)"
exec_asset_dir_regex = exec_dir_regex + r"/execution-asset/(?P<execution_asset_type>[-\w]+)"
exec_asset_regex = exec_asset_dir_regex + r"/(?P<file_name>[-\w]+)[.](?P<file_ext>[a-z0-9]+)$"
exec_metadata_dir_regex = exec_dir_regex + r'/execution-metadata/(?P<execution_metadata_type>[-\w]+)'
exec_metadata_regex = exec_metadata_dir_regex + r'/(?P<filename>[-\w]+)[.](?P<file_ext>[a-z0-9]*)$'
feature_dir_regex = exec_dir_regex + r'/feature'
feature_table_dir_regex = feature_dir_regex + r"/(?P<schema>[-\w]+)/(?P<target_table>[-\w]+)/(?P<feature_name>[-\w]+)"
feature_value_regex = feature_table_dir_regex + r"/(?P=feature_name)[.](?P<file_ext>[(csv|json)]*)$"
feature_asset_dir_regex =  feature_table_dir_regex + r"/asset/(?P<asset_table>[-\w]+)"
feature_asset_regex = feature_asset_dir_regex + r'/(?P<file_name>[A-Za-z0-9_-]+)[.](?P<file_ext>[a-z0-9]*)$'
asset_path_regex = (
    upload_root_regex
    + r"/asset/(?P<schema>[-\w]+)/(?P<asset_table>[-\w]*)/(?P<file_name>[-\w]+)[.](?P<file_ext>[a-z0-9]*)$"
)
table_regex = exec_dir_regex + r"/table/(?P<schema>[-\w]+)/(?P<table>[-\w]+)/(?P=table)[.](csv|json)$"


def is_execution_metadata_dir(path: Path) -> Optional[re.Match]:
    """

    Args:
      path: Path: 

    Returns:
        Regex match object from metadata_dir_regex.
    """
    return re.match(exec_metadata_dir_regex + '$', path.as_posix())


def is_execution_asset_dir(path: Path) -> Optional[re.Match]:
    """
    Args:
      path: Path: 

    Returns:

    """
    return re.match(exec_asset_dir_regex + '$', path.as_posix())


def is_feature_dir(path: Path) -> Optional[re.Match]:
    """

    Args:
      path: Path: 

    Returns:

    """
    return re.match(feature_table_dir_regex + '$', path.as_posix())


def is_feature_asset_dir(path: Path) -> Optional[re.Match]:
    """

    Args:
      path: Path: 

    Returns:

    """
    return re.match(feature_asset_dir_regex + '$', path.as_posix())


def upload_root(prefix: Path | str) -> Path:
    """

    Args:
      prefix: Path | str: 

    Returns:

    """
    path = Path(prefix) / "deriva-ml"
    path.mkdir(exist_ok=True, parents=True)
    return path


def execution_root(prefix: Path | str, exec_rid) -> Path:
    """

    Args:
      prefix: Path | str: 
      exec_rid: 

    Returns:

    """
    path = upload_root(prefix) / "execution" / exec_rid
    path.mkdir(exist_ok=True, parents=True)
    return path


def execution_asset_root(prefix: Path | str, exec_rid: str) -> Path:
    """

    Args:
      prefix: Path | str: 
      exec_rid: str: 

    Returns:

    """
    path = execution_root(prefix, exec_rid) / "execution-asset"
    path.mkdir(parents=True, exist_ok=True)
    return path


def execution_metadata_root(prefix: Path | str, exec_rid: str) -> Path:
    """

    Args:
      prefix: Path | str: 
      exec_rid: str: 

    Returns:

    """
    path = execution_root(prefix, exec_rid) / "execution-metadata"
    path.mkdir(parents=True, exist_ok=True)
    return path


def execution_asset_dir(prefix: Path | str, exec_rid: str, asset_type: str) -> Path:
    """Return the path to a directory in which to place execution assets of a specified type are to be uploaded.

    Args:
      prefix: Location of upload root directory
      asset_type: Type of execution asset
      exec_rid: RID of the execution asset
      prefix: Path | str: 
      exec_rid: str: 
      asset_type: str: 

    Returns:

    """
    path = execution_asset_root(prefix, exec_rid) / asset_type
    path.mkdir(parents=True, exist_ok=True)
    return path


def execution_metadata_dir(prefix: Path | str, exec_rid: str, metadata_type: str) -> Path:
    """Return the path to a directory in which to place execution metadata of a specified type are to be uploaded.

    Args:
        prefix: Location in which to locate this directory
        exec_rid: Execution rid to be associated with this metadata
        metadata_type: Controlled vocabulary term from vocabulary Metadata_Type

    Returns:
        Path to the metadata directory
    """
    path = execution_metadata_root(prefix, exec_rid) / metadata_type
    path.mkdir(parents=True, exist_ok=True)
    return path

def feature_root(prefix: Path | str, exec_rid: str) -> Path:
    """

    Args:
      prefix: Path | str: 
      exec_rid: str: 

    Returns:

    """
    path = execution_root(prefix, exec_rid) / 'feature'
    path.mkdir(parents=True, exist_ok=True)
    return path

def feature_dir(
    prefix: Path | str, exec_rid: str, schema: str, target_table: str, feature_name: str
) -> Path:
    """

    Args:
        prefix: Path | str:
        exec_rid: str:
        schema: str:
        target_table: str:
        feature_name: str:

    Returns:

    """
    path = feature_root(prefix, exec_rid) / schema / target_table / feature_name
    path.mkdir(parents=True, exist_ok=True)
    return path


def feature_value_path(
    prefix: Path | str, exec_rid: str, schema: str, target_table: str, feature_name: str
) -> Path:
    """Return the path to a CSV file in which to place feature values that are to be uploaded.  Values will either be
    scalar, references to controlled vocabulary (Terms) or references to assets.

    Args:
        prefix: Location of upload root directory
        exec_rid: RID of the execution to be associated with this feature.
        schema: Domain schema name
        target_table: Target table name for the feature.
        feature_name: Name of the feature.

    Returns:
        Path to CSV file in which to place feature values
    """
    return feature_dir(prefix, exec_rid, schema, target_table, feature_name) / f"{feature_name}.csv"


def feature_asset_dir(
    prefix: Path | str,
    exec_rid: str,
    schema: str,
    target_table: str,
    feature_name: str,
    asset_table: str,
) -> Path:
    """Return the path to a directory in which to place feature assets for a named feature are to be uploaded.

    Args:
        prefix: Location of upload root directory
        exec_rid: RID of the execution for the feature asset
        schema: Domain schema
        target_table: Name of the target table for the feature.
        feature_name: Name of the feature
        asset_table: Name of the asset table for the feature.

    Returns:
        Path to directory in which feature asset files are placed.
    """
    path = feature_dir(prefix, exec_rid, schema, target_table, feature_name) / 'asset' / asset_table

    path.mkdir(parents=True, exist_ok=True)
    return path


def asset_dir(prefix: Path | str, schema: str, asset_table: str) -> Path:
    """Return the path to a directory in which to place assets that are to be uploaded.

    Args:
        prefix: Location of upload root directory
        schema: Domain schema
        asset_table: Name of the asset table

    Returns:
        Path to the directory in which to place assets
    """
    path = upload_root(prefix) / 'asset' / schema / asset_table
    path.mkdir(parents=True, exist_ok=True)
    return path


def table_path(prefix: Path | str, schema: str, table: str) -> Path:
    """Return the path to a CSV file in which to place table values that are to be uploaded.

    Args:
        prefix: Location of upload root directory
        schema: Domain schema
        table: Name of the table to be uploaded.

    Returns:
        Path to the file in which to place table values that are to be uploaded.
    """
    path = upload_root(prefix) / 'table' / schema / table
    path.mkdir(parents=True, exist_ok=True)
    return path / f"{table}.csv"


bulk_upload_configuration = {
    "asset_mappings": [
        {
            # Upload  any files that may have been created by the program execution.  These are  in the
            # Execution_Metadata directory
            "column_map": {
                "MD5": "{md5}",
                "URL": "{URI}",
                "Length": "{file_size}",
                "Filename": "{file_name}",
                "Execution_Metadata_Type": "{execution_metadata_type_name}",
            },
            "file_pattern": exec_metadata_regex,
            "target_table": ["deriva-ml", "Execution_Metadata"],
            "checksum_types": ["sha256", "md5"],
            "hatrac_options": {"versioned_urls": True},
            "hatrac_templates": {
                "hatrac_uri": "/hatrac/execution_metadata/{md5}.{file_name}",
                "content-disposition": "filename*=UTF-8''{file_name}.{file_ext}",
            },
            "record_query_template": "/entity/{target_table}/MD5={md5}&Filename={file_name}",
            "metadata_query_templates": [
                "/attribute/deriva-ml:Execution_Metadata_Type/Name={execution_metadata_type}/execution_metadata_type_name:=Name"
            ],
        },
        {
            # Upload the contents of the Execution_Asset directory.
            "column_map": {
                "MD5": "{md5}",
                "URL": "{URI}",
                "Length": "{file_size}",
                "Filename": "{file_name}",
                "Execution_Asset_Type": "{execution_asset_type_name}",
            },
            "file_pattern": exec_asset_regex,
            "target_table": ["deriva-ml", "Execution_Asset"],
            "checksum_types": ["sha256", "md5"],
            "hatrac_options": {"versioned_urls": True},
            "hatrac_templates": {
                "hatrac_uri": "/hatrac/execution_asset/{md5}.{file_name}",
                "content-disposition": "filename*=UTF-8''{file_name}.{file_ext}",
            },
            "record_query_template": "/entity/{target_table}/MD5={md5}&Filename={file_name}",
            "metadata_query_templates": [
                "/attribute/deriva-ml:Execution_Asset_Type/Name={execution_asset_type}/execution_asset_type_name:=Name"
            ],
        },
        {
            # Upload the assets for a feature table.
            "column_map": {
                "MD5": "{md5}",
                "URL": "{URI}",
                "Length": "{file_size}",
                "Filename": "{file_name}",
            },
            "file_pattern": feature_asset_regex,  # Sets target_table, feature_name, asset_table
            "target_table": ["{schema}", "{asset_table}"],
            "checksum_types": ["sha256", "md5"],
            "hatrac_options": {"versioned_urls": True},
            "hatrac_templates": {
                "hatrac_uri": "/hatrac/{asset_table}/{md5}.{file_name}",
                "content-disposition": "filename*=UTF-8''{file_name}",
            },
            "record_query_template": "/entity/{target_table}/MD5={md5}&Filename={file_name}",
        },
        {
            # Upload assets into an asset table of an asset table.
            "column_map": {
                "MD5": "{md5}",
                "URL": "{URI}",
                "Length": "{file_size}",
                "Filename": "{file_name}",
            },
            "file_pattern": asset_path_regex,  # Sets schema, asset_table, file_name, file_ext
            "checksum_types": ["sha256", "md5"],
            "hatrac_options": {"versioned_urls": True},
            "hatrac_templates": {
                "hatrac_uri": "/hatrac/{asset_table}/{md5}.{file_name}",
                "content-disposition": "filename*=UTF-8''{file_name}.{file_ext}",
            },
            "target_table": ["{schema}", "{asset_table}"],
            "record_query_template": "/entity/{target_table}/MD5={md5}&Filename={file_name}",
        },
        #{
            #  Upload the records into a  table
         #   "asset_type": "skip",
         ##   "default_columns": ["RID", "RCB", "RMB", "RCT", "RMT"],
          #  "file_pattern": feature_value_regex,  # Sets schema, table,
          #  "ext_pattern": "^.*[.](?P<file_ext>json|csv)$",
          #  "target_table": ["{schema}", "{table}"],
       # },
        {
            #  Upload the records into a  table
            "asset_type": "table",
            "default_columns": ["RID", "RCB", "RMB", "RCT", "RMT"],
            "file_pattern": table_regex,  # Sets schema, table,
            "ext_pattern": "^.*[.](?P<file_ext>json|csv)$",
            "target_table": ["{schema}", "{table}"],
        },
    ],
    "version_update_url": "https://github.com/informatics-isi-edu/deriva-client",
    "version_compatibility": [[">=1.4.0", "<2.0.0"]],
}

def test_upload():
    """ """
    ead = execution_asset_dir('foo', 'my-rid', 'my-asset')
    emd = execution_metadata_dir('foo', 'my-rid', 'my-metadata')
    _fp = feature_value_path('foo', 'my-rid', 'my-schema', 'my-target', 'my-feature')
    fa = feature_asset_dir('foo', 'my-rid', 'my-schema', 'my-target', 'my-feature', 'my-asset')
    _tp = table_path('foo', 'my-schema', 'my-table')
    _ad = asset_dir('foo', 'my-schema', 'my-asset')
    _is_md = is_execution_metadata_dir(emd)
    _is_ea = is_execution_asset_dir(ead)
    _is_fa = is_feature_asset_dir(fa)



import logging
import sqlite3
from collections import defaultdict
from copy import copy
from csv import reader
from pathlib import Path
import tempfile
from typing import Any, Generator, Optional
from urllib.parse import urlparse

import pandas as pd
from deriva.core.ermrest_model import Model
from pydantic import validate_call

from .deriva_definitions import ML_SCHEMA, MLVocab, RID
from .deriva_ml_base import DerivaMLException


class DatasetBag(object):
    """DatasetBag is a class that manages a materialized bag.  It is created from a locally materialized BDBag for a
    dataset, which is created either by DerivaML.create_execution, or directly by calling DerivaML.download_dataset.

    As part of its initialization, this routine will create a sqlite database that has the contents of all the tables
    in the dataset.  In addition, any asset tables will the `Filename` column remapped to have the path of the local
    copy of the file. In addition, a local version of the ERMRest model that as used to generate the dataset is
    available.

       The sqllite database will not have any foreign key constraints applied, however, foreign-key relationships can be
    found by looking in the ERMrest model.  In addition, as sqllite doesn't support schema, Ermrest schema are added
    to the table name using the convention SchemaName:TableName.  Methods in DatasetBag that have table names as the
    argument will perform the appropriate name mappings.

    Attributes:
        dbase: A sqlite database that has the contents of all the tables in the dataset.

    Methods:
        get_table(table: str) -> Generator[tuple, None, None]
        get_table_as_dataframe(table: str) -> pd.DataFrame
        get_table_as_dict(table: str) -> Generator[dict[str, Any], None, None]
        list_tables() -> list[str]
    """

    dbase: Optional[sqlite3.Connection] = None
    _model: Optional[Model] = None
    _rids_loaded: dict[RID:Path] = {}
    _paths_loaded: set[Path] = set()

    @validate_call
    def __init__(self, dataset: Path | RID, dbase_dir: Optional[Path] = None) -> None:
        """
        Initialize a DatasetBag instance.

        Args:
            dataset: A path to a BDBag or a dataset, or the RID of an already loaded dataset.
            dbase_dir: Optional path as to where to place the sqllite database file.
        """

        if dataset in DatasetBag._rids_loaded:
            # We have already loaded this RID already, so just pick up the info and return.
            self.bag_path = DatasetBag._rids_loaded[dataset]
            self.dataset_rid = dataset
            self._ml_schema = ML_SCHEMA
        else:
            self.bag_path = Path(dataset) if isinstance(dataset, str) else dataset
            self.dataset_rid = dataset.name.replace("Dataset_", "")
            self._ml_schema = ML_SCHEMA
            if self.bag_path not in DatasetBag._paths_loaded:
                # This is the first time we have seen this bag, so we need to create a database for it and
                # load it up.
                self._create_database(dbase_dir)
                self._domain_schema = self._guess_domain_schema()
                self._ml_schema = ML_SCHEMA
                self._load_model()
                self._load_sqllite()
                DatasetBag._paths_loaded.add(self.bag_path)
                DatasetBag._rids_loaded[self.dataset_rid] = self.bag_path

                # If this is a nested dataset, we want to know what RIDs are in it.
                for d in self.list_dataset_children(recurse=True):
                    self._rids_loaded[d] = self.bag_path
        self._domain_schema = self._guess_domain_schema()
        self.dataset_table = DatasetBag._model.schemas[self._ml_schema].tables[
            "Dataset"
        ]

    @staticmethod
    def _guess_domain_schema():
        # Guess the domain schema name by eliminating all the "builtin" schema.
        return [
            s
            for s in DatasetBag._model.schemas
            if s not in ["deriva-ml", "public", "www"]
        ][0]

    def _create_database(self, dir_path: Optional[Path] = None) -> None:
        if not DatasetBag._model:
            DatasetBag._model = Model.fromfile(
                "file-system", self.bag_path / "data/schema.json"
            )
            dir_path = dir_path or Path(tempfile.mkdtemp())
            dbase_file = dir_path / "dataset.db"
            print(f"Creating database file: {dbase_file}")
            DatasetBag.dbase = sqlite3.connect(dbase_file)

    def _load_model(self):
        # Create a sqlite database schema that contains all the tables within the catalog from which the
        # BDBag was created.
        with DatasetBag.dbase:
            for t in DatasetBag._model.schemas[self._domain_schema].tables.values():
                DatasetBag.dbase.execute(t.sqlite3_ddl())
            for t in DatasetBag._model.schemas["deriva-ml"].tables.values():
                DatasetBag.dbase.execute(t.sqlite3_ddl())

    def _localize_asset_table(self) -> dict[str, str]:
        """Use the fetch.txt file in a bdbag to create a map from a URL to a local file path.

        Returns:
            Dictionary that maps a URL to a local file path.

        """
        fetch_map = {}
        try:
            with open(self.bag_path / "fetch.txt", newline="\n") as fetchfile:
                for row in fetchfile:
                    # Rows in fetch.text are tab seperated with URL filename.
                    fields = row.split("\t")
                    local_file = fields[2].replace('\n', "")
                    local_path =  f'{self.bag_path}/{local_file}'
                    fetch_map[urlparse(fields[0]).path] = local_path
        except FileNotFoundError:
            dataset_rid = self.bag_path.name.replace("Dataset_", "")
            logging.info(f"No downloaded assets in bag {dataset_rid}")
        return fetch_map

    def _is_asset(self, table_name: str) -> bool:
        """

        Args:
          table_name: str:

        Returns:

        """
        asset_columns = {"Filename", "URL", "Length", "MD5", "Description"}
        sname = (
            self._domain_schema
            if table_name in DatasetBag._model.schemas[self._domain_schema].tables
            else self._ml_schema
        )
        asset_table = DatasetBag._model.schemas[sname].tables[table_name]
        return asset_columns.issubset({c.name for c in asset_table.columns})

    @staticmethod
    def _localize_asset(
        o: list, indexes: tuple[int, int], asset_map: dict[str, str]
    ) -> tuple:
        """Given a list of column values for a table, replace the FileName column with the local file name based on
        the URL value.

        Args:
          o: List of values for each column in a table row.
          indexes: A tuple whose first element is the column index of the file name and whose second element
        is the index of the URL in an asset table.  Tuple is None if table is not an asset table.
          o: list:
          indexes: Optional[tuple[int: int]]:

        Returns:
          Tuple of updated column values.

        """
        if indexes:
            file_column, url_column = indexes
            o[file_column] = asset_map[o[url_column]] if o[url_column] else ""
        return tuple(o)

    def _load_sqllite(self) -> None:
        """Load a SQLite database from a bdbag.  THis is done by looking for all the CSV files in the bdbag directory.

        If the file is for an asset table, update the FileName column of the table to have the local file path for
        the materialized file.  Then load into the sqllite database.
        Note: none of the foreign key constraints are included in the database.
        """
        dpath = self.bag_path / "data"
        asset_map = self._localize_asset_table()  # Map of remote to local assets.

        # Find all the CSV files in the subdirectory and load each file into the database.
        for csv_file in Path(dpath).rglob("*.csv"):
            table = csv_file.stem
            schema = (
                self._domain_schema
                if table in DatasetBag._model.schemas[self._domain_schema].tables
                else self._ml_schema
            )

            with csv_file.open(newline="") as csvfile:
                csv_reader = reader(csvfile)
                column_names = next(csv_reader)

                # Determine which columns in the table has the Filename and the URL
                asset_indexes = (
                    (column_names.index("Filename"), column_names.index("URL"))
                    if self._is_asset(table)
                    else None
                )

                value_template = ",".join(
                    ["?"] * len(column_names)
                )  # SQL placeholder for row (?,?..)
                column_list = ",".join([f'"{c}"' for c in column_names])
                with DatasetBag.dbase:
                    object_table = (
                        DatasetBag._localize_asset(o, asset_indexes, asset_map)
                        for o in csv_reader
                    )
                    DatasetBag.dbase.executemany(
                        f'INSERT OR REPLACE INTO "{schema}:{table}" ({column_list}) VALUES ({value_template})',
                        object_table,
                    )

    @staticmethod
    def list_tables() -> list[str]:
        """List the names of the tables in the catalog

        Returns:
            A list of table names.  These names are all qualified with the Deriva schema name.
        """
        with DatasetBag.dbase:
            return [
                t[0]
                for t in DatasetBag.dbase.execute(
                    "SELECT name FROM sqlite_master WHERE type = 'table' ORDER BY name;"
                ).fetchall()
            ]

    @staticmethod
    def find_datasets() -> list[dict[str, Any]]:
        """Returns a list of currently available datasets.

        Returns:
             list of currently available datasets.
        """
        atable = next(
            DatasetBag._model.schemas[ML_SCHEMA]
            .tables[MLVocab.dataset_type]
            .find_associations()
        ).name

        # Get a list of all the dataset_type values associated with this dataset.
        datasets = []
        print(atable)
        ds_types = list(DatasetBag.get_table_as_dict(atable))
        print(ds_types)
        for dataset in DatasetBag.get_table_as_dict("Dataset"):
            my_types = [t for t in ds_types if t["Dataset"] == dataset["RID"]]
            datasets.append(
                dataset
                | {MLVocab.dataset_type: [ds[MLVocab.dataset_type] for ds in my_types]}
            )
        return datasets

    @validate_call
    def list_dataset_members(self, recurse: bool = False) -> defaultdict:
        """Return a list of entities associated with a specific dataset.

        Args:
           recurse:  (Default value = False)

        Returns:
            Dictionary of entities associated with a specific dataset.  Key is the table from which the elements
            were taken.
        """

        # Look at each of the element types that might be in the dataset and get the list of rid for them from
        # the appropriate association table.
        members = defaultdict(list)
        for assoc_table in self.dataset_table.find_associations():
            other_fkey = assoc_table.other_fkeys.pop()
            self_fkey = assoc_table.self_fkey
            target_table = other_fkey.pk_table
            member_table = assoc_table.table

            if (
                target_table.schema.name != self._domain_schema
                and target_table != self.dataset_table
            ):
                # Look at domain tables and nested datasets.
                continue
            if target_table == self.dataset_table:
                # find_assoc gives us the keys in the wrong position, so swap.
                self_fkey, other_fkey = other_fkey, self_fkey
            sql_target = DatasetBag._normalize_table_name(target_table.name)
            sql_member = DatasetBag._normalize_table_name(member_table.name)

            # Get the names of the columns that we are going to need for linking
            member_link = tuple(
                c.name for c in next(iter(other_fkey.column_map.items()))
            )

            with DatasetBag.dbase:
                sql_cmd = (
                    f'SELECT * FROM "{sql_member}" '
                    f'JOIN "{sql_target}" ON "{sql_member}".{member_link[0]} = "{sql_target}".{member_link[1]} '
                    f'WHERE "{self.dataset_rid}" = "{sql_member}".Dataset;'
                )
                target_entities = DatasetBag.dbase.execute(sql_cmd).fetchall()
                members[target_table.name].extend(target_entities)

            target_entities = []  # path.entities().fetch()
            members[target_table.name].extend(target_entities)
            if recurse and target_table.name == self.dataset_table:
                # Get the members for all the nested datasets and add to the member list.
                nested_datasets = [d["RID"] for d in target_entities]
                for ds in nested_datasets:
                    for k, v in DatasetBag.list_dataset_members(
                        ds, recurse=False
                    ).items():
                        members[k].extend(v)
        return members

    @validate_call
    def list_dataset_children(self, recurse: bool = False) -> list[RID]:
        """Given a dataset RID, return a list of RIDs of any nested datasets.

        Returns:
          list of RIDs of nested datasets.

        """
        return self._list_dataset_children(self.dataset_rid, recurse)

    def _list_dataset_children(self, dataset_rid, recurse: bool) -> list[RID]:
        ds_table = DatasetBag._normalize_table_name("Dataset_Dataset")
        with DatasetBag.dbase:
            nested = [
                r[0]
                for r in DatasetBag.dbase.execute(
                    f'SELECT Nested_Dataset FROM "{ds_table}" where Dataset = "{dataset_rid}"'
                ).fetchall()
            ]
        result = copy(nested)
        if recurse:
            for child in nested:
                result.extend(self._list_dataset_children(child, recurse))
        return result

    @staticmethod
    def _normalize_table_name(table: str) -> str:
        """Attempt to insert the schema into a table name if it's not provided.

        Args:
          table: str:

        Returns:
          table name with schema included.

        """
        sname = ""
        try:
            [sname, tname] = table.split(":")
        except ValueError:
            tname = table
            for sname, s in DatasetBag._model.schemas.items():
                if table in s.tables:
                    break
        try:
            _ = DatasetBag._model.schemas[sname].tables[tname]
            return f"{sname}:{tname}"
        except KeyError:
            raise DerivaMLException(f'Table name "{table}" does not exist.')

    @staticmethod
    def get_table(table: str) -> Generator[tuple, None, None]:
        """Retrieve the contents of the specified table. If schema is not provided as part of the table name,
        the method will attempt to locate the schema for the table.

        Args:
            table: return: A generator that yields tuples of column values.

        Returns:
          A generator that yields tuples of column values.

        """
        table_name = DatasetBag._normalize_table_name(table)
        result = DatasetBag.dbase.execute(f'SELECT * FROM "{table_name}"')
        while row := result.fetchone():
            yield row

    @staticmethod
    def get_table_as_dataframe(table: str) -> pd.DataFrame:
        """Retrieve the contents of the specified table as a dataframe.


        If schema is not provided as part of the table name,
        the method will attempt to locate the schema for the table.

        Args:
            table: Table to retrieve data from.

        Returns:
          A dataframe containing the contents of the specified table.
        """
        table_name = DatasetBag._normalize_table_name(table)
        print(table_name)
        return pd.read_sql(f'SELECT * FROM "{table_name}"', con=DatasetBag.dbase)

    @staticmethod
    def get_table_as_dict(table: str) -> Generator[dict[str, Any], None, None]:
        """Retrieve the contents of the specified table as a dictionary.

        Args:
            table: Table to retrieve data from. f schema is not provided as part of the table name,
                the method will attempt to locate the schema for the table.

        Returns:
          A generator producing dictionaries containing the contents of the specified table as name/value pairs.
        """
        table_name = DatasetBag._normalize_table_name(table)
        with DatasetBag.dbase:
            col_names = [
                c[1]
                for c in DatasetBag.dbase.execute(
                    f'PRAGMA table_info("{table_name}")'
                ).fetchall()
            ]
            result = DatasetBag.dbase.execute(f'SELECT * FROM "{table_name}"')
            while row := result.fetchone():
                yield dict(zip(col_names, row))

    @staticmethod
    def delete_database(bag_path: Path, schema: str):
        """

        Args:
          bag_path:
          schema:

        Returns:

        """
        dbase_path = Path(bag_path) / f"{schema}.db"
        dbase_path.unlink()



__all__ = [
    'DerivaML',
    'DerivaMLException',
    'FileUploadState',
    'ExecutionConfiguration',
    'Workflow',
    'DatasetBag',
    'ColumnDefinition',
    'TableDefinition',
    'BuiltinTypes',
    'UploadState',
    'MLVocab',
    'ExecMetadataVocab',
    'RID',
    'DerivaSystemColumns'
]

from .dataset_bag import DatasetBag
from .deriva_definitions import ColumnDefinition, TableDefinition, BuiltinTypes, UploadState, FileUploadState, RID
from .deriva_definitions import DerivaMLException, MLVocab, ExecMetadataVocab, DerivaSystemColumns
from .deriva_ml_base import DerivaML
from .execution_configuration import ExecutionConfiguration, Workflow



import argparse
import sys

from deriva.core import DerivaServer, ErmrestCatalog, get_credential
from deriva.core.ermrest_model import Model
from deriva.core.ermrest_model import builtin_types, Schema, Table, Column, ForeignKey, Key
from deriva.core.utils.core_utils import tag as chaise_tags

from deriva_ml import MLVocab
from deriva_ml.schema_setup.annotations import generate_annotation


def define_table_workflow(workflow_annotation: dict):
    return Table.define(
        "Workflow",
        column_defs=[
            Column.define("Name", builtin_types.text),
            Column.define("Description", builtin_types.markdown),
            Column.define("URL", builtin_types.ermrest_uri),
            Column.define("Checksum", builtin_types.text),
            Column.define("Version", builtin_types.text),
        ],
        annotations=workflow_annotation,
    )

def define_table_dataset(dataset_annotation: dict = None):
    return Table.define(
        tname="Dataset",
        column_defs=[
            Column.define("Version", builtin_types.text, default="1.0.0", nullok=True),
            Column.define("Description", builtin_types.markdown)],
        annotations=dataset_annotation if dataset_annotation is not None else {},
    )


def define_table_execution(sname: str, execution_annotation: dict):
    table_def = Table.define(
        "Execution",
        column_defs=[
            Column.define("Workflow", builtin_types.text),
            Column.define("Description", builtin_types.markdown),
            Column.define("Duration", builtin_types.text),
            Column.define("Status", builtin_types.text),
            Column.define("Status_Detail", builtin_types.text),
        ],
        fkey_defs=[ForeignKey.define(["Workflow"], sname, "Workflow", ["RID"])],
        annotations=execution_annotation,
    )
    return table_def


def define_asset_execution_metadata(sname: str, execution_metadata_annotation: dict):
    return Table.define_asset(
        sname=sname,
        tname="Execution_Metadata",
        column_defs=[Column.define("Execution", builtin_types.markdown)],
        hatrac_template="/hatrac/metadata/{{MD5}}.{{Filename}}",
        fkey_defs=[ForeignKey.define(["Execution"], sname, "Execution", ["RID"])],
        annotations=execution_metadata_annotation,
    )


def define_asset_execution_asset(sname: str, execution_asset_annotation: dict):
    table_def = Table.define_asset(
        sname=sname,
        tname="Execution_Asset",
        hatrac_template="/hatrac/execution_asset/{{MD5}}.{{Filename}}",
        annotations=execution_asset_annotation,
    )
    return table_def

def create_www_schema(model: Model):
    """
    Set up a new schema and tables to hold web-page like content.  The tables include a page table, and a asset
    table that can have images that are referred to by the web page.  Pages are written using markdown.
    :return:
    """
    if model.schemas.get('www'):
        model.schemas['www'].drop(cascade=True)
    www_schema = model.create_schema(
        Schema.define(
        'www', comment='Schema for tables that will be displayed as web content')
    )
    www_schema.create_table(
        Table.define(
            'Page',
            column_defs=[
                Column.define('Title', builtin_types.text, nullok=False, comment='Unique title for the page'),
                Column.define('Content', builtin_types.markdown, comment='Content of the page in markdown')
            ],
            key_defs=[Key.define(['Title'])],
            annotations={
                chaise_tags.table_display: {'detailed': {'hide_column_headers': True, 'collapse_toc_panel': True}
                                            },
                chaise_tags.visible_foreign_keys: {'detailed': {}},
                chaise_tags.visible_columns: {'detailed': ['Content']}}

        )
    )

    return www_schema

def create_ml_schema(model: Model, schema_name: str = 'deriva-ml', project_name: str = None):
    ml_catalog: ErmrestCatalog = model.catalog

    if model.schemas.get(schema_name):
        model.schemas[schema_name].drop(cascade=True)
    # get annotations
    annotations = generate_annotation(ml_catalog.catalog_id, schema_name)
    model.annotations.update(annotations['catalog_annotation'])
    client_annotation = {
        "tag:misd.isi.edu,2015:display": {"name": "Users"},
        "tag:isrd.isi.edu,2016:table-display": {"row_name": {"row_markdown_pattern": "{{{Full_Name}}}"}},
        "tag:isrd.isi.edu,2016:visible-columns": {"compact": ["Full_Name", "Display_Name", "Email", "ID"]}
    }
    model.schemas['public'].tables['ERMrest_Client'].annotations.update(client_annotation)
    model.apply()

    schema = model.create_schema(Schema.define(schema_name, annotations=annotations['schema_annotation']))
    project_name = project_name or schema_name
    # Workflow
    schema.create_table(Table.define_vocabulary("Feature_Name", f'{project_name}:{{RID}}'))

    workflow_table = schema.create_table(define_table_workflow(annotations["workflow_annotation"]))
    workflow_table.create_reference(schema.create_table(
        Table.define_vocabulary(MLVocab.workflow_type, f'{schema_name}:{{RID}}')))

    execution_table = schema.create_table(define_table_execution(schema_name, annotations["execution_annotation"]))

    dataset_table = schema.create_table(define_table_dataset(annotations["dataset_annotation"]))
    dataset_type = schema.create_table(
        Table.define_vocabulary(MLVocab.dataset_type, f'{project_name}:{{RID}}'))
    schema.create_table(
        Table.define_association(associates=[("Dataset", dataset_table), (MLVocab.dataset_type, dataset_type)])
    )
    schema.create_table(
        Table.define_association(associates=[("Dataset", dataset_table), ("Execution", execution_table)])
    )

    # Nested datasets.
    schema.create_table(
        Table.define_association(associates=[("Dataset", dataset_table), ("Nested_Dataset", dataset_table)])
    )

    # Execution Metadata
    execution_metadata_table = schema.create_table(
        define_asset_execution_metadata(schema.name, annotations["execution_metadata_annotation"]))
    execution_metadata_table.create_reference(
        schema.create_table(
            Table.define_vocabulary("Execution_Metadata_Type", f'{project_name}:{{RID}}')))
    schema.create_table(
        Table.define_association([("Execution_Metadata", execution_metadata_table), ("Execution", execution_table)]))

    # Execution Asset
    execution_asset_table = schema.create_table(
        define_asset_execution_asset(schema.name, annotations["execution_asset_annotation"])
    )
    execution_asset_table.create_reference(
        schema.create_table(
            Table.define_vocabulary("Execution_Asset_Type", f'{project_name}:{{RID}}')))
    schema.create_table(
        Table.define_association([("Execution_Asset", execution_asset_table), ("Execution", execution_table)]))

    create_www_schema(model)
    initialize_ml_schema(model, schema_name)

def initialize_ml_schema(model: Model, schema_name: str = 'deriva-ml'):
    catalog = model.catalog
    execution_metadata_type = catalog.getPathBuilder().schemas[schema_name].tables['Execution_Metadata_Type']
    execution_metadata_type.insert([{'Name': 'Execution_Config',
                                     'Description': "Configuration File for execution metadata"},
                                    {'Name': "Runtime_Env",
                                     'Description': "Information about the execution environment"}],
                                   defaults={'ID', 'URI'})


def main():
    scheme = "https"
    parser = argparse.ArgumentParser()
    parser.add_argument("--hostname", type=str, required=True)
    parser.add_argument("--schema_name", type=str, required=True)
    parser.add_argument("--catalog_id", type=str, required=True)
    parser.add_argument("--curie_prefix", type=str, required=True)
    args = parser.parse_args()
    credentials = get_credential(args.hostname)
    server = DerivaServer(scheme, args.hostname, credentials)
    model = server.connect_ermrest(args.catalog_id).getCatalogModel()
    create_ml_schema(model, args.schema_name)


if __name__ == "__main__":
    sys.exit(main())



# Tests for the datapath module.
#
# Environment variables:
#  DERIVA_PY_TEST_HOSTNAME: hostname of the test server
#  DERIVA_PY_TEST_CREDENTIAL: user credential, if none, it will attempt to get credentail for given hostname
#  DERIVA_PY_TEST_VERBOSE: set for verbose logging output to stdout
import logging
import os
import sys
import unittest

from deriva.core import DerivaServer, ErmrestCatalog, get_credential
from typing import Optional

from src.deriva_ml import DerivaML, DerivaMLException, RID, ColumnDefinition, BuiltinTypes
from src.deriva_ml import create_ml_schema
from test_catalog import create_domain_schema, populate_test_catalog
from src.deriva_ml import ExecutionConfiguration

try:
    from pandas import DataFrame

    HAS_PANDAS = True
except ImportError:
    HAS_PANDAS = False

SNAME_DOMAIN = 'ml-test'

hostname = os.getenv("DERIVA_PY_TEST_HOSTNAME")
logger = logging.getLogger(__name__)
if os.getenv("DERIVA_PY_TEST_VERBOSE"):
    logger.setLevel(logging.DEBUG)
    logger.addHandler(logging.StreamHandler())


test_catalog: Optional[ErmrestCatalog] = None


def setUpModule():
    global test_catalog
    logger.debug("setUpModule begin")
    credential = os.getenv("DERIVA_PY_TEST_CREDENTIAL") or get_credential(hostname)
    server = DerivaServer('https', hostname, credentials=credential)
    test_catalog = server.create_ermrest_catalog()
    model = test_catalog.getCatalogModel()
    try:
        create_ml_schema(model)
        create_domain_schema(model, SNAME_DOMAIN)
    except Exception:
        # on failure, delete catalog and re-raise exception
        test_catalog.delete_ermrest_catalog(really=True)
        raise
    logger.debug("setUpModule  done")


def tearDownModule():
    logger.debug("tearDownModule begin")
    test_catalog.delete_ermrest_catalog(really=True)
    logger.debug("tearDownModule done")


@unittest.skipUnless(hostname, "Test host not specified")
class TestVocabulary(unittest.TestCase):

    def setUp(self):
        self.ml_instance = DerivaML(hostname, test_catalog.catalog_id, SNAME_DOMAIN, None, None, "1")
        self.domain_schema = self.ml_instance.model.schemas[SNAME_DOMAIN]
        self.model = self.ml_instance.model

    def tearDown(self):
        pass

    def test_find_vocabularies(self):
        populate_test_catalog(self.ml_instance, SNAME_DOMAIN)
        self.assertIn("Dataset_Type", [v.name for v in self.ml_instance.find_vocabularies()])

    def test_create_vocabulary(self):
        populate_test_catalog(self.ml_instance, SNAME_DOMAIN)
        self.ml_instance.create_vocabulary("CV1", "A vocab")
        self.assertTrue(self.model.schemas[self.ml_instance.domain_schema].tables["CV1"])

    def test_add_term(self):
        populate_test_catalog(self.ml_instance, SNAME_DOMAIN)
        self.ml_instance.create_vocabulary("CV1", "A vocab")
        self.assertEqual(len(self.ml_instance.list_vocabulary_terms("CV1")), 0)
        term = self.ml_instance.add_term("CV1", "T1", description="A vocab")
        self.assertEqual(len(self.ml_instance.list_vocabulary_terms("CV1")), 1)
        self.assertEqual(term.name, self.ml_instance.lookup_term("CV1", "T1").name)

        # Check for redudent terms.
        with self.assertRaises(DerivaMLException) as context:
            self.ml_instance.add_term("CV1", "T1", description="A vocab", exists_ok=False)
        self.assertEqual("T1", self.ml_instance.add_term("CV1", "T1", description="A vocab").name)


@unittest.skipUnless(hostname, "Test host not specified")
class TestFeatures(unittest.TestCase):
    def setUp(self):
        self.ml_instance = DerivaML(hostname, test_catalog.catalog_id, SNAME_DOMAIN, "", "", "1")
        self.domain_schema = self.ml_instance.model.schemas[SNAME_DOMAIN]
        self.model = self.ml_instance.model

    def test_create_feature(self):
        populate_test_catalog(self.ml_instance, SNAME_DOMAIN)
        self.ml_instance.create_vocabulary("FeatureValue", "A vocab")
        self.ml_instance.add_term("FeatureValue", "V1", description="A Feature Vale")

        a = self.ml_instance.create_asset("TestAsset", comment="A asset")

        self.ml_instance.create_feature("Feature1",
                                        "Image",
                                        terms=["FeatureValue"],
                                        assets=[a],
                                        metadata=[ColumnDefinition(name='TestCol', type=BuiltinTypes.int2)])
        self.assertIn("Execution_Image_Feature1",
                      [f.name for f in self.ml_instance.find_features("Image")])

    def test_add_feature(self):
        self.test_create_feature()
        TestFeature = self.ml_instance.feature_record_class("Image", "Feature1")
        # Create the name for this feature and then create the feature.
        # Get some images to attach the feature value to.
        domain_path = self.ml_instance.catalog.getPathBuilder().schemas[SNAME_DOMAIN]
        image_rids = [i['RID'] for i in domain_path.tables['Image'].entities().fetch()]
        asset_rid = domain_path.tables["TestAsset"].insert(
            [{'Name': "foo", 'URL': "foo/bar", 'Length': 2, 'MD5': 4}])[0]['RID']
        # Get an execution RID.
        ml_path = self.ml_instance.catalog.getPathBuilder().schemas['deriva-ml']
        self.ml_instance.add_term("Workflow_Type", "TestWorkflow", description="A workflow")
        workflow_rid = \
            ml_path.tables['Workflow'].insert([{'Name': "Test Workflow", 'Workflow_Type': "TestWorkflow"}])[0]['RID']
        execution_rid = \
            ml_path.tables['Execution'].insert([{'Description': "Test execution", 'Workflow': workflow_rid}])[0]['RID']
        # Now create a list of features using the feature creation class returned by create_feature.
        feature_list = [TestFeature(
            Image=i,
            Execution=execution_rid,
            FeatureValue="V1",
            TestAsset=asset_rid,
            TestCol=23) for i in image_rids]
        self.ml_instance.add_features(feature_list)
        features = self.ml_instance.list_feature_values("Image", "Feature1")
        self.assertEqual(len(features), len(image_rids))


class TestExecution(unittest.TestCase):
    def setUp(self):
        self.ml_instance = DerivaML(hostname, test_catalog.catalog_id, SNAME_DOMAIN, None, None, "1")
        self.domain_schema = self.ml_instance.model.schemas[SNAME_DOMAIN]
        self.model = self.ml_instance.model
        self.files = os.path.dirname(__file__) + '/files'

    def test_upload_configuration(self):
        populate_test_catalog(self.ml_instance, SNAME_DOMAIN)
        config_file = self.files + "/test-workflow-1.json"
        config = ExecutionConfiguration.load_configuration(config_file)
        rid = self.ml_instance.upload_execution_configuration(config)
        self.assertEqual(rid, self.ml_instance.retrieve_rid(rid)['RID'])

    def test_execution_1(self):
        populate_test_catalog(self.ml_instance, SNAME_DOMAIN)
        exec_config = ExecutionConfiguration.load_configuration(self.files + "/test-workflow-1.json")
        configuration_record = self.ml_instance.initialize_execution(configuration=exec_config)
        with self.ml_instance.execution(configuration=configuration_record) as exec:
            output_dir = self.ml_instance.execution_assets_path / "testoutput"
            output_dir.mkdir(parents=True, exist_ok=True)
            with open(output_dir / "test.txt", "w+") as f:
                f.write("Hello there\n")
        upload_status = self.ml_instance.upload_execution(configuration=configuration_record)
        e = (list(self.ml_instance.catalog.getPathBuilder().deriva_ml.Execution.entities().fetch()))[0]
        self.assertEqual(e['Status'], "Completed")



class TestDataset(unittest.TestCase):
    def setUp(self):
        self.ml_instance = DerivaML(hostname, test_catalog.catalog_id, SNAME_DOMAIN, None, None, "1")
        self.domain_schema = self.ml_instance.model.schemas[SNAME_DOMAIN]
        self.model = self.ml_instance.model

    def test_add_element_type(self):
        populate_test_catalog(self.ml_instance, SNAME_DOMAIN)
        self.ml_instance.add_dataset_element_type("Subject")
        self.assertEqual(len(list(self.ml_instance.dataset_table.find_associations())), 2)

    def test_create_dataset(self) -> RID:
        populate_test_catalog(self.ml_instance, SNAME_DOMAIN)
        self.ml_instance.add_dataset_element_type("Subject")
        type_rid = self.ml_instance.add_term("Dataset_Type", "TestSet", description="A test")
        dataset_rid = self.ml_instance.create_dataset(type_rid.name, description="A Dataset")
        self.assertEqual(len(self.ml_instance.find_datasets()), 1)
        return dataset_rid

    def test_add_dataset_members(self):
        dataset_rid = self.test_create_dataset()
        subject_rids = [i['RID'] for i in self.ml_instance.catalog.getPathBuilder().schemas[SNAME_DOMAIN].tables[
            'Subject'].entities().fetch()]
        self.ml_instance.add_dataset_members(dataset_rid=dataset_rid, members=subject_rids)
        self.assertEqual(len(self.ml_instance.list_dataset_members(dataset_rid)["Subject"]), len(subject_rids))


if __name__ == '__main__':
    sys.exit(unittest.main())



from typing import Any

from deriva.core.ermrest_model import Model
from deriva.core.utils.core_utils import tag as deriva_tags

from deriva_ml.dataset import Dataset


def dataset_visible_columns(model: Model) -> dict[str, Any]:
    dataset_table = model.schemas['deriva-ml'].tables['Dataset']
    rcb_name = next(
        [fk.name[0].name, fk.name[1]] for fk in dataset_table.foreign_keys if fk.name[1] == "Dataset_RCB_fkey")
    rmb_name = next(
        [fk.name[0].name, fk.name[1]] for fk in dataset_table.foreign_keys if fk.name[1] == "Dataset_RMB_fkey")
    return {
        "*": [
            "RID",
            "Description",
            {"display": {
                "markdown_pattern": "[Annotate Dataset](https://www.eye-ai.org/apps/grading-interface/main?dataset_rid={{{RID}}}){: .btn}"
            },
                "markdown_name": "Annotation App"
            },
            rcb_name,
            rmb_name
        ],
        'detailed': [
            "RID",
            "Description",
            {'source': [{"inbound": ['deriva-ml', 'Dataset_Dataset_Type_Dataset_fkey']},
                        {"outbound": ['deriva-ml', 'Dataset_Dataset_Type_Dataset_Type_fkey']}, 'RID'],
             'markdown_name': 'Dataset Types'},
            {"display": {
                "markdown_pattern": "[Annotate Dataset](https://www.eye-ai.org/apps/grading-interface/main?dataset_rid={{{RID}}}){: .btn}"
            },
                "markdown_name": "Annotation App"
            },
            rcb_name,
            rmb_name
        ],
        'filter': {
            'and': [
                {'source': 'RID'},
                {'source': 'Description'},
                {'source': [{"inbound": ['deriva-ml', 'Dataset_Dataset_Type_Dataset_fkey']},
                            {"outbound": ['deriva-ml', 'Dataset_Dataset_Type_Dataset_Type_fkey']}, 'RID'],
                 'markdown_name': 'Dataset Types'},
                {'source': [{'outbound': rcb_name}, 'RID'], 'markdown_name': 'Created By'},
                {'source': [{'outbound': rmb_name}, 'RID'], 'markdown_name': 'Modified By'},
            ]
        }
    }


def dataset_visible_fkeys(model: Model) -> dict[str, Any]:
    def fkey_name(fk):
        return [fk.name[0].name, fk.name[1]]

    dataset_table = model.schemas['deriva-ml'].tables['Dataset']

    source_list = [
        {"source": [
            {"inbound": fkey_name(fkey.self_fkey)},
            {"outbound": fkey_name(other_fkey := fkey.other_fkeys.pop())},
            "RID"
        ],
            "markdown_name": other_fkey.pk_table.name
        }
        for fkey in dataset_table.find_associations(max_arity=3, pure=False)
    ]
    return {'detailed': source_list}


def generate_dataset_annotations(model: Model) -> dict[str, Any]:
    ds = Dataset(model)
    return {
        deriva_tags.export_fragment_definitions: {'dataset_export_outputs': ds.export_outputs()},
        deriva_tags.visible_columns: dataset_visible_columns(model),
        deriva_tags.visible_foreign_keys: dataset_visible_fkeys(model),
        deriva_tags.export_2019: {
            'detailed': {
                'templates': [
                    {
                        'type': 'BAG',
                        'outputs': [{'fragment_key': 'dataset_export_outputs'}],
                        'displayname': 'BDBag Download',
                        'bag_idempotent': True,
                        'postprocessors': [
                            {
                                'processor': 'identifier',
                                'processor_params': {
                                    'test': False,
                                    'env_column_map': {'Dataset_RID': '{RID}@{snaptime}',
                                                       'Description': '{Description}'}
                                }
                            }
                        ]
                    },
                    {
                        'type': 'BAG',
                        'outputs': [{'fragment_key': 'dataset_export_outputs'}],
                        'displayname': 'BDBag to Cloud',
                        'bag_idempotent': True,
                        'postprocessors': [
                            {
                                'processor': 'cloud_upload',
                                'processor_params': {'acl': 'public-read', 'target_url': 's3://eye-ai-shared/'}
                            },
                            {
                                'processor': 'identifier',
                                'processor_params': {
                                    'test': False,
                                    'env_column_map': {'Dataset_RID': '{RID}@{snaptime}',
                                                       'Description': '{Description}'}
                                }
                            }
                        ]
                    }
                ]
            }
        }
    }





from deriva_ml import DatasetBag
from deriva_ml.dataset import Dataset
from deriva_ml.demo_catalog import create_demo_catalog, DemoML
test_catalog = create_demo_catalog('dev.eye-ai.org', 'demo-schema', create_features=True, create_datasets=True)
ml_instance = DemoML('dev.eye-ai.org', test_catalog.catalog_id)
bag = ml_instance.download_dataset_bag('3QY')
ds = DatasetBag(bag[0])

ds = DatasetBag(bag[1])